import pandas as pd

from woodwork.accessor_utils import _is_dask_series, _is_koalas_series
from woodwork.exceptions import TypeConversionError
from woodwork.type_sys.utils import _get_specified_ltype_params
from woodwork.utils import (
    _infer_datetime_format,
    _reformat_to_latlong,
    camel_to_snake,
    import_or_none
)

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


class ClassNameDescriptor(object):
    """Descriptor to convert a class's name from camelcase to snakecase"""

    def __get__(self, instance, class_):
        return camel_to_snake(class_.__name__)


class LogicalTypeMetaClass(type):

    def __repr__(cls):
        return cls.__name__


class LogicalType(object, metaclass=LogicalTypeMetaClass):
    """Base class for all other Logical Types"""
    type_string = ClassNameDescriptor()
    primary_dtype = 'string'
    backup_dtype = None
    standard_tags = set()

    def __eq__(self, other, deep=False):
        return isinstance(other, self.__class__) and _get_specified_ltype_params(other) == _get_specified_ltype_params(self)

    def __str__(self):
        return str(self.__class__)

    @classmethod
    def _get_valid_dtype(cls, series_type):
        """Return the dtype that is considered valid for a series with the given logical_type"""
        if ks and series_type == ks.Series and cls.backup_dtype:
            return cls.backup_dtype
        else:
            return cls.primary_dtype

    def transform(self, series):
        """Converts the series dtype to match the logical type's if it is different."""
        new_dtype = self._get_valid_dtype(type(series))
        if new_dtype != str(series.dtype):
            # Update the underlying series
            try:
                series = series.astype(new_dtype)
            except (TypeError, ValueError):
                raise TypeConversionError(series, new_dtype, type(self))
        return series


class Address(LogicalType):
    """Represents Logical Types that contain address values.

    Examples:
        .. code-block:: python

            ['1 Miller Drive, New York, NY 12345', '1 Berkeley Street, Boston, MA 67891']
            ['26387 Russell Hill, Dallas, TX 34521', '54305 Oxford Street, Seattle, WA 95132']
    """
    primary_dtype = 'string'


class Age(LogicalType):
    """Represents Logical Types that contain non-negative numbers indicating a person's age. Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [15, 22, 45]
            [30, 62, 87]
    """
    primary_dtype = 'int64'
    standard_tags = {'numeric'}


class AgeNullable(LogicalType):
    """Represents Logical Types that contain non-negative numbers indicating a person's age. Has 'numeric' as a standard tag. May also contain null values.

    Examples:
        .. code-block:: python

            [np.nan, 22, 45]
            [30, 62, np.nan]
    """
    primary_dtype = 'Int64'
    standard_tags = {'numeric'}


class Boolean(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.

    Examples:
        .. code-block:: python

            [True, False, True]
            [0, 1, 1]
    """
    primary_dtype = 'bool'


class BooleanNullable(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.
    May also contain null values.

    Examples:
        .. code-block:: python

            [True, False, None]
            [0, 1, 1]
    """
    primary_dtype = 'boolean'


class Categorical(LogicalType):
    """Represents Logical Types that contain unordered discrete values that fall
    into one of a set of possible values. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["red", "green", "blue"]
            ["produce", "dairy", "bakery"]
            [3, 1, 2]
    """
    primary_dtype = 'category'
    backup_dtype = 'string'
    standard_tags = {'category'}

    def __init__(self, encoding=None):
        # encoding dict(str -> int)
        # user can specify the encoding to use downstream
        pass


class CountryCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard country code to represent countries.
    ISO 3166-1 (countries) are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["AU", "US", "UA"]
            ["GB", "NZ", "DE"]
    """
    primary_dtype = 'category'
    backup_dtype = 'string'
    standard_tags = {'category'}


class Datetime(LogicalType):
    """Represents Logical Types that contain date and time information.

    Args:
        datetime_format (str): Desired datetime format for data

    Examples:
        .. code-block:: python

            ["2020-09-10",
             "2020-01-10 00:00:00",
             "01/01/2000 08:30"]
    """
    primary_dtype = 'datetime64[ns]'
    datetime_format = None

    def __init__(self, datetime_format=None):
        self.datetime_format = datetime_format

    def transform(self, series):
        """Converts the series data to a formatted datetime. Datetime format will be inferred if datetime_format is None."""
        new_dtype = self._get_valid_dtype(type(series))
        if new_dtype != str(series.dtype):
            sample_length = len(series.index) // 10 or len(series.index)
            self.datetime_format = self.datetime_format or _infer_datetime_format(series, n=sample_length)
            try:
                if _is_dask_series(series):
                    name = series.name
                    series = dd.to_datetime(series, format=self.datetime_format)
                    series.name = name
                elif _is_koalas_series(series):
                    series = ks.Series(ks.to_datetime(series.to_numpy(),
                                                      format=self.datetime_format),
                                       name=series.name)
                else:
                    series = pd.to_datetime(series, format=self.datetime_format)
            except (TypeError, ValueError):
                raise TypeConversionError(series, new_dtype, type(self))
        return super().transform(series)


class Double(LogicalType):
    """Represents Logical Types that contain positive and negative numbers, some of
    which include a fractional component. Includes zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [1.2, 100.4, 3.5]
            [-15.34, 100, 58.3]
    """
    primary_dtype = 'float64'
    standard_tags = {'numeric'}


class Integer(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, 0]
            [-54, 73, 11]
    """
    primary_dtype = 'int64'
    standard_tags = {'numeric'}


class IntegerNullable(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0). May contain null
    values. Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, np.nan]
            [-54, 73, 11]
    """
    primary_dtype = 'Int64'
    standard_tags = {'numeric'}


class EmailAddress(LogicalType):
    """Represents Logical Types that contain email address values.

    Examples:
        .. code-block:: python

            ["john.smith@example.com",
             "support@example.com",
             "team@example.com"]
    """
    primary_dtype = 'string'


class Filepath(LogicalType):
    """Represents Logical Types that specify locations of directories and files
    in a file system.

    Examples:
        .. code-block:: python

            ["/usr/local/bin",
             "/Users/john.smith/dev/index.html",
             "/tmp"]
    """
    primary_dtype = 'string'


class PersonFullName(LogicalType):
    """Represents Logical Types that may contain first, middle and last names,
    including honorifics and suffixes.

    Examples:
        .. code-block:: python

            ["Mr. John Doe, Jr.",
             "Doe, Mrs. Jane",
             "James Brown"]
    """
    primary_dtype = 'string'


class IPAddress(LogicalType):
    """Represents Logical Types that contain IP addresses, including both
    IPv4 and IPv6 addresses.

    Examples:
        .. code-block:: python

            ["172.16.254.1",
             "192.0.0.0",
             "2001:0db8:0000:0000:0000:ff00:0042:8329"]
    """
    primary_dtype = 'string'


class LatLong(LogicalType):
    """Represents Logical Types that contain latitude and longitude values in decimal degrees.

    Note:
        LatLong values will be stored with the object dtype as a
        tuple of floats (or a list of floats for Koalas DataFrames)
        and must contain only two values.

        Null latitude or longitude values will be stored as np.nan, and
        a fully null LatLong (np.nan, np.nan) will be stored as just a
        single nan.

    Examples:
        .. code-block:: python

            [(33.670914, -117.841501),
             (40.423599, -86.921162),
             (-45.031705, nan)]
    """
    primary_dtype = 'object'

    def transform(self, series):
        """Formats a series to be a tuple (or list for Koalas) of two floats."""
        if _is_dask_series(series):
            name = series.name
            meta = (series, tuple([float, float]))
            series = series.apply(_reformat_to_latlong, meta=meta)
            series.name = name
        elif _is_koalas_series(series):
            formatted_series = series.to_pandas().apply(_reformat_to_latlong, use_list=True)
            series = ks.from_pandas(formatted_series)
        else:
            series = series.apply(_reformat_to_latlong)

        return super().transform(series)


class NaturalLanguage(LogicalType):
    """Represents Logical Types that contain text or characters representing
    natural human language

    Examples:
        .. code-block:: python

            ["This is a short sentence.",
             "I like to eat pizza!",
             "When will humans go to mars?"]
    """
    primary_dtype = 'string'


class Unknown(LogicalType):
    """Represents Logical Types that cannot be inferred as a specific Logical Type. It is assumed to contain string data.

    Examples:
        .. code-block:: python

            ["ax23n9ck23l",
             "1,28&*_%*&&xejc",
             "xnmvz@@Dcmeods-0"]

    """
    primary_dtype = 'string'


class Ordinal(LogicalType):
    """Represents Logical Types that contain ordered discrete values.
    Has 'category' as a standard tag.

    Args:
        order (list or tuple): An list or tuple specifying the order of the ordinal
            values from low to high. The underlying series cannot contain values that
            are not present in the order values.

    Examples:
        .. code-block:: python

            ["first", "second", "third"]
            ["bronze", "silver", "gold"]
    """
    primary_dtype = 'category'
    backup_dtype = 'string'
    standard_tags = {'category'}

    def __init__(self, order=None):
        self.order = order

    def _validate_data(self, series):
        """Make sure order values are properly defined and confirm the supplied series
        does not contain any values that are not in the specified order values"""
        if self.order is None:
            raise TypeError("Must use an Ordinal instance with order values defined")
        elif not isinstance(self.order, (list, tuple)):
            raise TypeError("Order values must be specified in a list or tuple")
        if len(self.order) != len(set(self.order)):
            raise ValueError("Order values cannot contain duplicates")

        if isinstance(series, pd.Series):
            missing_order_vals = set(series.dropna().values).difference(self.order)
            if missing_order_vals:
                error_msg = f'Ordinal column {series.name} contains values that are not present ' \
                    f'in the order values provided: {sorted(list(missing_order_vals))}'
                raise ValueError(error_msg)

    def transform(self, series):
        """Validates the series and converts the dtype to match the logical type's if it is different."""
        self._validate_data(series)
        return super().transform(series)


class PhoneNumber(LogicalType):
    """Represents Logical Types that contain numeric digits and characters
    representing a phone number

    Examples:
        .. code-block:: python

            ["1-(555)-123-5495",
             "+1-555-123-5495",
             "5551235495"]
    """
    primary_dtype = 'string'


class SubRegionCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard sub-region code to
    represent a portion of a larger geographic region. ISO 3166-2 (sub-regions)
    codes are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["US-CO", "US-MA", "US-CA"]
            ["AU-NSW", "AU-TAS", "AU-QLD"]
    """
    primary_dtype = 'category'
    backup_dtype = 'string'
    standard_tags = {'category'}


class Timedelta(LogicalType):
    """Represents Logical Types that contain values specifying a duration of time

    Examples:
        .. code-block:: python

            [pd.Timedelta('1 days 00:00:00'),
             pd.Timedelta('-1 days +23:40:00'),
             pd.Timedelta('4 days 12:00:00')]
    """
    primary_dtype = 'timedelta64[ns]'


class URL(LogicalType):
    """Represents Logical Types that contain URLs, which may include protocol, hostname
    and file name

    Examples:
        .. code-block:: python

            ["http://google.com",
             "https://example.com/index.html",
             "example.com"]
    """
    primary_dtype = 'string'


class PostalCode(LogicalType):
    """Represents Logical Types that contain a series of postal codes for
    representing a group of addresses. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["90210"
             "60018-0123",
             "SW1A"]
    """
    primary_dtype = 'category'
    backup_dtype = 'string'
    standard_tags = {'category'}



import pandas as pd
import pandas.api.types as pdtypes

import woodwork as ww
from woodwork.type_sys.utils import _is_categorical_series, col_is_datetime


def categorical_func(series):
    if pdtypes.is_categorical_dtype(series.dtype):
        return True

    if pdtypes.is_string_dtype(series.dtype) and not col_is_datetime(series):
        categorical_threshold = ww.config.get_option('categorical_threshold')

        return _is_categorical_series(series, categorical_threshold)

    if pdtypes.is_float_dtype(series.dtype) or pdtypes.is_integer_dtype(series.dtype):
        numeric_categorical_threshold = ww.config.get_option('numeric_categorical_threshold')
        if numeric_categorical_threshold is not None:
            return _is_categorical_series(series, numeric_categorical_threshold)
        else:
            return False

    return False


def integer_func(series):
    if integer_nullable_func(series) and not series.isnull().any():
        return True
    return False


def integer_nullable_func(series):
    if pdtypes.is_integer_dtype(series.dtype):
        threshold = ww.config.get_option('numeric_categorical_threshold')
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True

    return False


def double_func(series):
    if pdtypes.is_float_dtype(series.dtype):
        threshold = ww.config.get_option('numeric_categorical_threshold')
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True

    return False


def boolean_func(series):
    if boolean_nullable_func(series) and not series.isnull().any():
        return True
    return False


def boolean_nullable_func(series):
    if pdtypes.is_bool_dtype(series.dtype) and not pdtypes.is_categorical_dtype(series.dtype):
        return True
    return False


def datetime_func(series):
    if col_is_datetime(series):
        return True
    return False


def timedelta_func(series):
    if pdtypes.is_timedelta64_dtype(series.dtype):
        return True
    return False


def email_address_func(series: pd.Series) -> bool:
    regex = ww.config.get_option('email_inference_regex')

    # Includes a check for object dtypes
    if not pdtypes.is_string_dtype(series.dtype):
        return False

    try:
        series_match_method = series.str.match
    except (AttributeError, TypeError):
        # This can happen either when the inferred dtype for a series is not
        # compatible with the pandas string API (AttributeError) *or* when the
        # inferred dtype is not compatible with the string API `match` method
        # (TypeError)
        return False
    matches = series_match_method(pat=regex)

    return matches.sum() == matches.count()



import re

import pandas as pd
import pytest

from woodwork.accessor_utils import _is_dask_series, _is_koalas_series
from woodwork.exceptions import TypeConversionError
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    LatLong,
    Ordinal
)


def test_logical_eq():
    assert Boolean == Boolean
    assert Boolean() == Boolean()
    assert Categorical != Boolean
    assert Datetime != Datetime()
    assert Datetime() == Datetime(datetime_format=None)
    assert Datetime() != Datetime(datetime_format='%Y-%m-%d')


def test_logical_repr():
    assert isinstance(repr(Boolean), str)
    assert repr(Boolean) == 'Boolean'
    assert isinstance(repr(Categorical), str)
    assert repr(Categorical) == 'Categorical'


def test_instantiated_type_str():
    assert str(Categorical()) == 'Categorical'
    assert str(Boolean()) == 'Boolean'


def test_ordinal_order_errors():
    series = pd.Series([1, 2, 3]).astype('category')

    with pytest.raises(TypeError, match='Order values must be specified in a list or tuple'):
        Ordinal(order='not_valid').transform(series)

    with pytest.raises(ValueError, match='Order values cannot contain duplicates'):
        Ordinal(order=['a', 'b', 'b']).transform(series)


def test_ordinal_init_with_order():
    order = ['bronze', 'silver', 'gold']
    ordinal_from_list = Ordinal(order=order)
    assert ordinal_from_list.order == order

    order = ('bronze', 'silver', 'gold')
    ordinal_from_tuple = Ordinal(order=order)
    assert ordinal_from_tuple.order == order


def test_get_valid_dtype(sample_series):
    valid_dtype = Categorical._get_valid_dtype(type(sample_series))
    if _is_koalas_series(sample_series):
        assert valid_dtype == 'string'
    else:
        assert valid_dtype == 'category'

    valid_dtype = Boolean._get_valid_dtype(type(sample_series))
    assert valid_dtype == 'bool'


def test_latlong_transform(latlong_df):
    df_type = str(type(latlong_df))
    dask = 'dask' in df_type
    koalas = 'koalas' in df_type
    nan = float('nan')

    expected_data = {
        'tuple_ints': [(1.0, 2.0), (3.0, 4.0)],
        'tuple_strings': [(1.0, 2.0), (3.0, 4.0)],
        'string_tuple': [(1.0, 2.0), (3.0, 4.0)],
        'bracketless_string_tuple': [(1.0, 2.0), (3.0, 4.0)],
        'list_strings': [(1.0, 2.0), (3.0, 4.0)],
        'combo_tuple_types': [(1.0, 2.0), (3.0, 4.0)],
        'null_value': [nan, (3.0, 4.0)],
        'null_latitude': [(nan, 2.0), (3.0, 4.0)],
        'both_null': [nan, (3.0, 4.0)],
    }

    latlong = LatLong()
    for column in latlong_df:
        series = latlong_df[column]
        actual = latlong.transform(series)

        if dask:
            actual = actual.compute()
        elif koalas:
            actual = actual.to_pandas()

        actual = actual.apply(pd.Series)
        series = pd.Series(expected_data[column])
        expected = series.apply(pd.Series)
        assert actual.equals(expected)


def test_datetime_transform(datetimes):
    datetime = Datetime()
    for series in datetimes:
        assert str(series.dtype) == 'object'
        transform = datetime.transform(series)
        assert str(transform.dtype) == 'datetime64[ns]'
        assert datetime.datetime_format is not None


def test_datetime_conversion_error(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail('Dask does not show error until compute is made.')

    dtype = str(sample_series.dtype)
    match = f'Error converting datatype for sample_series from type {dtype} to type datetime64[ns]. '
    match += 'Please confirm the underlying data is consistent with logical type Datetime.'
    with pytest.raises(TypeConversionError, match=re.escape(match)):
        Datetime().transform(sample_series)


def test_ordinal_transform(sample_series):
    series_type = str(type(sample_series))
    dask = 'dask' in series_type
    koalas = 'koalas' in series_type

    if dask or koalas:
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not supported')

    ordinal_incomplete_order = Ordinal(order=['a', 'b'])
    error_msg = re.escape("Ordinal column sample_series contains values that are not "
                          "present in the order values provided: ['c']")

    with pytest.raises(ValueError, match=error_msg):
        ordinal_incomplete_order.transform(sample_series)



import copy
from collections.abc import Hashable

import pandas as pd

import woodwork as ww
from woodwork.column_schema import ColumnSchema
from woodwork.exceptions import ColumnNotPresentError
from woodwork.type_sys.utils import _get_ltype_class
from woodwork.utils import _convert_input_to_set


class TableSchema(object):
    def __init__(self, column_names,
                 logical_types,
                 name=None,
                 index=None,
                 time_index=None,
                 semantic_tags=None,
                 table_metadata=None,
                 column_metadata=None,
                 use_standard_tags=False,
                 column_descriptions=None,
                 column_origins=None,
                 validate=True):
        """Create TableSchema

        Args:
            column_names (list, set): The columns present in the TableSchema.
            logical_types (dict[str -> LogicalType]): Dictionary mapping column names in
                the TableSchema to the LogicalType for the column. All columns present in the
                TableSchema must be present in the logical_types dictionary.
            name (str, optional): Name used to identify the TableSchema.
            index (str, optional): Name of the index column.
            time_index (str, optional): Name of the time index column.
            semantic_tags (dict, optional): Dictionary mapping column names in the TableSchema to the
                semantic tags for the column. The keys in the dictionary should be strings
                that correspond to columns in the TableSchema. There are two options for
                specifying the dictionary values:
                (str): If only one semantic tag is being set, a single string can be used as a value.
                (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
                used as the value.
                Semantic tags will be set to an empty set for any column not included in the
                dictionary.
            table_metadata (dict[str -> json serializable], optional): Dictionary containing extra metadata for the TableSchema. The dictionary must contain
                data types that are JSON serializable such as string, integers, and floats. DataFrame and Series types are not supported.
            column_metadata (dict[str -> dict[str -> json serializable]], optional): Dictionary mapping column names
                to that column's metadata dictionary.
            use_standard_tags (bool, dict[str -> bool], optional): Determines whether standard semantic tags will be
                added to columns based on the specified logical type for the column.
                If a single boolean is supplied, will apply the same use_standard_tags value to all columns.
                A dictionary can be used to specify ``use_standard_tags`` values for individual columns.
                Unspecified columns will use the default value. Defaults to False.
            column_descriptions (dict[str -> str], optional): Dictionary mapping column names to column descriptions.
            column_origins (str, dict[str -> str], optional): Origin of each column. If a string is supplied, it is
                used as the origin for all columns. A dictionary can be used to set origins for individual columns.
            validate (bool, optional): Whether parameter validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """
        if validate:
            # Check that inputs are valid
            _validate_params(column_names, name, index, time_index, logical_types, table_metadata,
                             column_metadata, semantic_tags, column_descriptions, column_origins,
                             use_standard_tags)

        self._name = name
        self._metadata = table_metadata or {}

        # use_standard_tags should be a dictionary mapping each column to its boolean
        if isinstance(use_standard_tags, bool):
            use_standard_tags = {col_name: use_standard_tags for col_name in column_names}
        else:
            use_standard_tags = {**{col_name: False for col_name in column_names}, **use_standard_tags}

        # Infer logical types and create columns
        self.columns = self._create_columns(column_names,
                                            logical_types,
                                            semantic_tags,
                                            use_standard_tags,
                                            column_descriptions,
                                            column_origins,
                                            column_metadata,
                                            validate)
        if index is not None:
            self.set_index(index, validate=validate)

        if time_index is not None:
            self.set_time_index(time_index, validate=validate)

    def __eq__(self, other, deep=True):
        if self.name != other.name:
            return False
        if self.index != other.index:
            return False
        if self.time_index != other.time_index:
            return False
        if set(self.columns.keys()) != set(other.columns.keys()):
            return False
        for col_name in self.columns:
            if not self.columns[col_name].__eq__(other.columns[col_name], deep=deep):
                return False
        if deep and self.metadata != other.metadata:
            return False

        return True

    def __repr__(self):
        """A string representation of a TableSchema containing typing information."""
        return repr(self._get_typing_info())

    def _repr_html_(self):
        """An HTML representation of a TableSchema for IPython.display in Jupyter Notebooks
        containing typing information and a preview of the data."""
        return self._get_typing_info().to_html()

    @property
    def types(self):
        """DataFrame containing the physical dtypes, logical types and semantic
        tags for the TableSchema."""
        return self._get_typing_info()

    def _get_typing_info(self):
        """Creates a DataFrame that contains the typing information for a TableSchema."""
        typing_info = {}
        for col_name, col in self.columns.items():
            types = [col.logical_type, str(list(col.semantic_tags))]
            typing_info[col_name] = types

        columns = ['Logical Type', 'Semantic Tag(s)']

        df = pd.DataFrame.from_dict(typing_info,
                                    orient='index',
                                    columns=columns,
                                    dtype="object")
        df.index.name = 'Column'
        return df

    @property
    def name(self):
        """Name of schema"""
        return self._name

    @name.setter
    def name(self, name):
        """Set name of schema"""
        if name:
            _check_name(name)
        self._name = name

    @property
    def metadata(self):
        """Metadata of the table"""
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        """Set table metadata"""
        if metadata:
            _check_table_metadata(metadata)
        self._metadata = metadata or {}

    @property
    def logical_types(self):
        """A dictionary containing logical types for each column"""
        return {col_name: col.logical_type for col_name, col in self.columns.items()}

    @property
    def semantic_tags(self):
        """A dictionary containing semantic tags for each column"""
        return {col_name: col.semantic_tags for col_name, col in self.columns.items()}

    @property
    def index(self):
        """The index column for the table"""
        for col_name, column in self.columns.items():
            if 'index' in column.semantic_tags:
                return col_name
        return None

    @property
    def time_index(self):
        """The time index column for the table"""
        for col_name, column in self.columns.items():
            if 'time_index' in column.semantic_tags:
                return col_name
        return None

    @property
    def use_standard_tags(self):
        return {col_name: col.use_standard_tags for col_name, col in self.columns.items()}

    def set_types(self, logical_types=None, semantic_tags=None, retain_index_tags=True):
        """Update the logical type and semantic tags for any columns names in the provided types dictionaries,
        updating the TableSchema at those columns.

        Args:
            logical_types (dict[str -> LogicalType], optional): A dictionary defining the new logical types for the
                specified columns.
            semantic_tags (dict[str -> str/list/set], optional): A dictionary defining the new semantic_tags for the
                specified columns.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will replace all semantic tags any time a column's
                semantic tags or logical type changes. Defaults to True.
        """
        logical_types = logical_types or {}
        _check_logical_types(self.columns.keys(), logical_types, require_all_cols=False)

        semantic_tags = semantic_tags or {}
        _check_semantic_tags(self.columns.keys(), semantic_tags)

        for col_name in (logical_types.keys() | semantic_tags.keys()):
            original_tags = self.semantic_tags[col_name]

            # Update Logical Type for the TableSchema, getting new semantic tags
            new_logical_type = logical_types.get(col_name)
            if new_logical_type is not None:
                self.columns[col_name].logical_type = new_logical_type

            # Set new semantic tags, removing existing tags
            new_semantic_tags = semantic_tags.get(col_name)
            if new_semantic_tags is None:
                self.columns[col_name]._reset_semantic_tags()
            else:
                self.columns[col_name]._set_semantic_tags(new_semantic_tags)
                _validate_not_setting_index_tags(self.semantic_tags[col_name], col_name)

            if retain_index_tags and 'index' in original_tags:
                self._set_index_tags(col_name)
            if retain_index_tags and 'time_index' in original_tags:
                self._set_time_index_tags(col_name)

    def add_semantic_tags(self, semantic_tags):
        """Adds specified semantic tags to columns, updating the Woodwork typing information.
        Will retain any previously set values.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be added to the column's semantic tags
        """
        _check_semantic_tags(self.columns.keys(), semantic_tags)
        for col_name, tags_to_add in semantic_tags.items():
            tags_to_add = _convert_input_to_set(tags_to_add)
            _validate_not_setting_index_tags(tags_to_add, col_name)
            self.columns[col_name]._add_semantic_tags(tags_to_add, col_name)

    def remove_semantic_tags(self, semantic_tags):
        """Remove the semantic tags for any column names in the provided semantic_tags
        dictionary, updating the Woodwork typing information. Including `index` or `time_index`
        tags will set the Woodwork index or time index to None for the DataFrame.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be removed from the column's semantic tags
        """
        _check_semantic_tags(self.columns.keys(), semantic_tags)
        for col_name, tags_to_remove in semantic_tags.items():
            standard_tags = self.logical_types[col_name].standard_tags
            tags_to_remove = _convert_input_to_set(tags_to_remove)
            original_tags = self.semantic_tags[col_name].copy()

            self.columns[col_name]._remove_semantic_tags(tags_to_remove,
                                                         col_name)

            # If the index is removed, reinsert any standard tags not explicitly removed
            if self.use_standard_tags[col_name] and 'index' in original_tags and 'index' not in self.columns[col_name].semantic_tags:
                standard_tags_removed = tags_to_remove.intersection(standard_tags)
                standard_tags_to_reinsert = standard_tags.difference(standard_tags_removed)
                self.columns[col_name].semantic_tags = self.semantic_tags[col_name].union(standard_tags_to_reinsert)

    def reset_semantic_tags(self, columns=None, retain_index_tags=False):
        """Reset the semantic tags for the specified columns to the default values.
        The default values will be either an empty set or a set of the standard tags
        based on the column logical type, controlled by the use_standard_tags property on the table.
        Column names can be provided as a single string, a list of strings or a set of strings.
        If columns is not specified, tags will be reset for all columns.

        Args:
            columns (str/list/set, optional): The columns for which the semantic tags should be reset.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will clear all semantic tags. Defaults to
                False.
        """
        columns = _convert_input_to_set(columns, "columns")
        cols_not_found = sorted(list(columns.difference(set(self.columns.keys()))))
        if cols_not_found:
            raise ColumnNotPresentError(cols_not_found)
        if not columns:
            columns = self.columns.keys()

        for col_name in columns:
            original_tags = self.semantic_tags[col_name]
            self.columns[col_name]._reset_semantic_tags()

            if retain_index_tags and 'index' in original_tags:
                self._set_index_tags(col_name)
            if retain_index_tags and 'time_index' in original_tags:
                self._set_time_index_tags(col_name)

    def _create_columns(self,
                        column_names,
                        logical_types,
                        semantic_tags,
                        use_standard_tags,
                        column_descriptions,
                        column_origins,
                        column_metadata,
                        validate):
        """Create a dictionary with column names as keys and new column dictionaries holding
        each column's typing information as values."""
        columns = {}
        for name in column_names:
            semantic_tags_for_col = _convert_input_to_set((semantic_tags or {}).get(name),
                                                          error_language=f'semantic_tags for {name}', validate=validate)
            if validate:
                _validate_not_setting_index_tags(semantic_tags_for_col, name)
            description = (column_descriptions or {}).get(name)
            origin = column_origins if isinstance(column_origins, str) else (column_origins or {}).get(name)
            metadata_for_col = (column_metadata or {}).get(name)

            columns[name] = ColumnSchema(logical_type=logical_types.get(name),
                                         semantic_tags=semantic_tags_for_col,
                                         use_standard_tags=use_standard_tags.get(name),
                                         description=description,
                                         origin=origin,
                                         metadata=metadata_for_col,
                                         validate=validate)
        return columns

    def set_index(self, new_index, validate=True):
        """Sets the index. Handles setting a new index, updating the index, or removing the index.

        Args:
            new_index (str): Name of the new index column. Must be present in the TableSchema.
                If None, will remove the index.
        """
        old_index = self.index
        if old_index is not None:
            self.remove_semantic_tags({old_index: 'index'})
        if new_index is not None:
            if validate:
                _check_index(self.columns.keys(), new_index)

                if 'time_index' in self.columns[new_index].semantic_tags:
                    info = f'"{new_index}" is already set as the time index. '
                    info += 'A time index cannot also be the index.'
                    raise ValueError(info)

            self._set_index_tags(new_index)

    def set_time_index(self, new_time_index, validate=True):
        """Set the time index. Adds the 'time_index' semantic tag to the column and
        clears the tag from any previously set index column

        Args:
            new_time_index (str): The name of the column to set as the time index.
                If None, will remove the time_index.
        """
        old_time_index = self.time_index
        if old_time_index is not None:
            self.remove_semantic_tags({old_time_index: 'time_index'})
        if new_time_index is not None:
            if validate:
                _check_time_index(self.columns.keys(), new_time_index, self.logical_types.get(new_time_index))

                if 'index' in self.columns[new_time_index].semantic_tags:
                    info = f'"{new_time_index}" is already set as the index. '
                    info += 'An index cannot also be the time index.'
                    raise ValueError(info)

            self._set_time_index_tags(new_time_index)

    def rename(self, columns):
        """Renames columns in a TableSchema

        Args:
            columns (dict[str -> str]): A dictionary mapping current column names to new column names.

        Returns:
            woodwork.TableSchema: TableSchema with the specified columns renamed.
        """
        if not isinstance(columns, dict):
            raise TypeError('columns must be a dictionary')

        for old_name, new_name in columns.items():
            if old_name not in self.columns:
                raise ColumnNotPresentError(f"Column to rename must be present. {old_name} cannot be found.")
            if new_name in self.columns and new_name not in columns.keys():
                raise ValueError(f"The column {new_name} is already present. Please choose another name to rename {old_name} to or also rename {old_name}.")

        if len(columns) != len(set(columns.values())):
            raise ValueError('New columns names must be unique from one another.')

        new_schema = copy.deepcopy(self)

        cols_to_update = {}
        for old_name, new_name in columns.items():
            col = new_schema.columns.pop(old_name)
            cols_to_update[new_name] = col

        new_schema.columns.update(cols_to_update)
        return new_schema

    def _set_index_tags(self, index):
        """Updates the semantic tags of the index by removing any standard tags
        before adding the 'index' tag."""
        column = self.columns[index]

        standard_tags = column.logical_type.standard_tags
        new_tags = column.semantic_tags.difference(standard_tags)
        new_tags.add('index')

        self.columns[index].semantic_tags = new_tags

    def _set_time_index_tags(self, time_index):
        self.columns[time_index].semantic_tags.add('time_index')

    def _filter_cols(self, include=None, exclude=None, col_names=False):
        """Return list of columns filtered with any of: semantic tags, LogicalTypes, column names

        Args:
            include (str or LogicalType or list[str or LogicalType]): parameter or list of parameters to
                filter columns by. Can be Logical Types or Semantic Tags. Columns that match will be
                included in the returned list of columns.
            exclude (str or LogicalType or list[str or LogicalType]): parameter or list of parameters to
                filter columns by. Can be Logical Types or Semantic Tags. Columns that match will be
                excluded from the returned list of columns.

            col_names (bool): Specifies whether to filter columns by name. Defaults to False.

        Returns:
            List[str] of column names that fit into filter.
        """
        assert not (include and exclude), "Cannot specify both include and exclude"
        if include and not isinstance(include, list):
            include = [include]
        elif exclude and not isinstance(exclude, list):
            exclude = [exclude]

        if include is not None:
            selectors = include
        elif exclude is not None:
            selectors = exclude

        ltypes_used = set()
        ltypes_in_schema = {type(col.logical_type) for col in self.columns.values()}

        tags_used = set()
        tags_in_schema = {tag for col in self.columns.values() for tag in col.semantic_tags}

        col_name_matches = set()

        for selector in selectors:
            # Determine if the selector is a registered, uninstantiated LogicalType
            maybe_ltype = selector
            if isinstance(selector, str):
                # Convert possible string to LogicalType - unregistered LogicalTypes return None
                maybe_ltype = ww.type_system.str_to_logical_type(selector, raise_error=False)
            # Get the class - unregistered LogicalTypes return LogicalTypeMetaClass
            maybe_ltype_class = _get_ltype_class(maybe_ltype)

            if maybe_ltype_class in ww.type_system.registered_types:
                if maybe_ltype not in ww.type_system.registered_types:
                    raise TypeError(f"Invalid selector used in include: {maybe_ltype} cannot be instantiated")
                if maybe_ltype in ltypes_in_schema:
                    ltypes_used.add(maybe_ltype)
            elif maybe_ltype_class == ww.logical_types.LogicalType.__class__:
                raise TypeError(f"Specified LogicalType selector {maybe_ltype} is not registered in Woodwork's type system.")

            # Hashability as a proxy for whether a selector is possibly a semantic tag or column name
            if not isinstance(selector, Hashable):
                raise TypeError(f"Invalid selector used in include: {selector} must be a "
                                "string, uninstantiated and registered LogicalType, or valid column name")
            # Determine if the selector is a semantic tag
            if selector in tags_in_schema:
                tags_used.add(selector)
            # Determine if the selector is a column name
            if col_names and selector in self.columns:
                col_name_matches.add(selector)

        cols_to_return = []
        cols_seen = set()
        for col_name, col in self.columns.items():
            is_match = (type(col.logical_type) in ltypes_used or
                        col.semantic_tags.intersection(tags_used) or
                        col_name in col_name_matches)
            if include is not None and is_match and col_name not in cols_seen:
                cols_to_return.append(col_name)
                cols_seen.add(col_name)
            elif exclude is not None and not is_match and col_name not in cols_seen:
                cols_to_return.append(col_name)
                cols_seen.add(col_name)

        return cols_to_return

    def _get_subset_schema(self, subset_cols):
        """Creates a new TableSchema with specified columns, retaining typing information.

        Args:
            subset_cols (list[str]): subset of columns from which to create the new TableSchema
        Returns:
            TableSchema: New TableSchema with attributes from original TableSchema
        """
        new_logical_types = {}
        new_semantic_tags = {}
        new_column_descriptions = {}
        new_column_origins = {}
        new_column_metadata = {}
        for col_name in subset_cols:
            col = col = self.columns[col_name]

            new_logical_types[col_name] = col.logical_type
            new_semantic_tags[col_name] = col.semantic_tags
            new_column_descriptions[col_name] = col.description
            new_column_origins[col_name] = col.origin
            new_column_metadata[col_name] = col.metadata

        new_index = self.index if self.index in subset_cols else None
        new_time_index = self.time_index if self.time_index in subset_cols else None
        if new_index is not None:
            new_semantic_tags[new_index] = new_semantic_tags[new_index].difference({'index'})
        if new_time_index is not None:
            new_semantic_tags[new_time_index] = new_semantic_tags[new_time_index].difference({'time_index'})

        return TableSchema(subset_cols,
                           new_logical_types,
                           name=self.name,
                           index=new_index,
                           time_index=new_time_index,
                           semantic_tags=copy.deepcopy(new_semantic_tags),
                           use_standard_tags=self.use_standard_tags.copy(),
                           table_metadata=copy.deepcopy(self.metadata),
                           column_metadata=copy.deepcopy(new_column_metadata),
                           column_descriptions=new_column_descriptions,
                           column_origins=new_column_origins,
                           validate=False)


def _validate_params(column_names, name, index, time_index, logical_types,
                     table_metadata, column_metadata, semantic_tags, column_descriptions,
                     column_origins, use_standard_tags):
    """Check that values supplied during TableSchema initialization are valid"""
    _check_column_names(column_names)
    _check_use_standard_tags(column_names, use_standard_tags)
    if name:
        _check_name(name)
    if index is not None:
        _check_index(column_names, index)
    if logical_types:
        _check_logical_types(column_names, logical_types)
    if table_metadata:
        _check_table_metadata(table_metadata)
    if column_metadata:
        _check_column_metadata(column_names, column_metadata)
    if time_index is not None:
        _check_time_index(column_names, time_index, logical_types.get(time_index))
    if semantic_tags:
        _check_semantic_tags(column_names, semantic_tags)
    if column_descriptions:
        _check_column_descriptions(column_names, column_descriptions)
    if column_origins:
        _check_column_origins(column_names, column_origins)


def _check_name(name):
    if not isinstance(name, str):
        raise TypeError('Table name must be a string')


def _check_column_names(column_names):
    if not isinstance(column_names, (list, set)):
        raise TypeError('Column names must be a list or set')

    if len(column_names) != len(set(column_names)):
        raise IndexError('TableSchema cannot contain duplicate columns names')


def _check_index(column_names, index):
    if index not in column_names:
        # User specifies an index that is not in the list of column names
        raise ColumnNotPresentError(f'Specified index column `{index}` not found in TableSchema.')


def _check_time_index(column_names, time_index, logical_type):
    if time_index not in column_names:
        raise ColumnNotPresentError(f'Specified time index column `{time_index}` not found in TableSchema')
    ltype_class = _get_ltype_class(logical_type)

    if not (ltype_class == ww.logical_types.Datetime or 'numeric' in ltype_class.standard_tags):
        raise TypeError('Time index column must be a Datetime or numeric column.')


def _check_logical_types(column_names, logical_types, require_all_cols=True):
    if not isinstance(logical_types, dict):
        raise TypeError('logical_types must be a dictionary')
    cols_in_ltypes = set(logical_types.keys())
    cols_in_schema = set(column_names)

    cols_not_found_in_schema = cols_in_ltypes.difference(cols_in_schema)
    if cols_not_found_in_schema:
        raise ColumnNotPresentError('logical_types contains columns that are not present in '
                                    f'TableSchema: {sorted(list(cols_not_found_in_schema))}')
    cols_not_found_in_ltypes = cols_in_schema.difference(cols_in_ltypes)
    if cols_not_found_in_ltypes and require_all_cols:
        raise ColumnNotPresentError(f'logical_types is missing columns that are present in '
                                    f'TableSchema: {sorted(list(cols_not_found_in_ltypes))}')

    for col_name, logical_type in logical_types.items():
        if _get_ltype_class(logical_type) not in ww.type_system.registered_types:
            raise TypeError("Logical Types must be of the LogicalType class "
                            "and registered in Woodwork's type system. "
                            f"{logical_type} does not meet that criteria.")


def _check_semantic_tags(column_names, semantic_tags):
    if not isinstance(semantic_tags, dict):
        raise TypeError('semantic_tags must be a dictionary')
    cols_not_found = set(semantic_tags.keys()).difference(set(column_names))
    if cols_not_found:
        raise ColumnNotPresentError('semantic_tags contains columns that are not present in '
                                    f'TableSchema: {sorted(list(cols_not_found))}')

    for col_name, col_tags in semantic_tags.items():
        if not isinstance(col_tags, (str, list, set)):
            raise TypeError(f"semantic_tags for {col_name} must be a string, set or list")


def _check_column_descriptions(column_names, column_descriptions):
    if not isinstance(column_descriptions, dict):
        raise TypeError('column_descriptions must be a dictionary')
    cols_not_found = set(column_descriptions.keys()).difference(set(column_names))
    if cols_not_found:
        raise ColumnNotPresentError('column_descriptions contains columns that are not present in '
                                    f'TableSchema: {sorted(list(cols_not_found))}')


def _check_column_origins(column_names, column_origins):
    if not isinstance(column_origins, (dict, str)):
        raise TypeError('column_origins must be a dictionary or a string')

    if isinstance(column_origins, dict):
        cols_not_found = set(column_origins.keys()).difference(set(column_names))
        if cols_not_found:
            raise ColumnNotPresentError('column_origins contains columns that are not present in '
                                        f'TableSchema: {sorted(list(cols_not_found))}')


def _check_table_metadata(table_metadata):
    if not isinstance(table_metadata, dict):
        raise TypeError('Table metadata must be a dictionary.')


def _check_column_metadata(column_names, column_metadata):
    if not isinstance(column_metadata, dict):
        raise TypeError('Column metadata must be a dictionary.')
    cols_not_found = set(column_metadata.keys()).difference(set(column_names))
    if cols_not_found:
        raise ColumnNotPresentError('column_metadata contains columns that are not present in '
                                    f'TableSchema: {sorted(list(cols_not_found))}')


def _check_use_standard_tags(column_names, use_standard_tags):
    if not isinstance(use_standard_tags, (dict, bool)):
        raise TypeError('use_standard_tags must be a dictionary or a boolean')
    if isinstance(use_standard_tags, dict):
        cols_not_found = set(use_standard_tags.keys()).difference(set(column_names))
        if cols_not_found:
            raise ColumnNotPresentError('use_standard_tags contains columns that are not present in '
                                        f'TableSchema: {sorted(list(cols_not_found))}')

        for col_name, use_standard_tags_for_col in use_standard_tags.items():
            if not isinstance(use_standard_tags_for_col, bool):
                raise TypeError(f"use_standard_tags for column {col_name} must be a boolean")


def _validate_not_setting_index_tags(semantic_tags, col_name):
    """Verify user has not supplied tags that cannot be set directly"""
    if 'index' in semantic_tags:
        raise ValueError(f"Cannot add 'index' tag directly for column {col_name}. To set a column as the index, "
                         "use DataFrame.ww.set_index() instead.")
    if 'time_index' in semantic_tags:
        raise ValueError(f"Cannot add 'time_index' tag directly for column {col_name}. To set a column as the time index, "
                         "use DataFrame.ww.set_time_index() instead.")



from os import path

from setuptools import find_packages, setup

dirname = path.abspath(path.dirname(__file__))
with open(path.join(dirname, 'README.md')) as f:
    long_description = f.read()

extras_require = {'dask': open('dask-requirements.txt').readlines(),
                  'koalas': open('koalas-requirements.txt').readlines(),
                  'update_checker': ['alteryx-open-src-update-checker >= 2.0.0']}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

setup(
    name='woodwork',
    author='Alteryx, Inc.',
    author_email='support@featurelabs.com',
    license='BSD 3-clause',
    version='0.6.0',
    description='a two-dimensional data object with labeled axes and typing information',
    url='https://github.com/alteryx/woodwork/',
    classifiers=[
         'Development Status :: 3 - Alpha',
         'Intended Audience :: Developers',
         'Programming Language :: Python :: 3',
         'Programming Language :: Python :: 3.7',
         'Programming Language :: Python :: 3.8',
         'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(),
    install_requires=open('requirements.txt').readlines(),
    tests_require=open('test-requirements.txt').readlines(),
    python_requires='>=3.7, <4',
    extras_require=extras_require,
    keywords='data science machine learning typing',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'woodwork = woodwork.__main__:cli'
        ]
    },
    long_description=long_description,
    long_description_content_type='text/markdown'
)



import os

import pandas as pd
import pytest
from mock import patch

import woodwork as ww
from woodwork.serialize import save_orc_file


def test_read_file_errors_no_content_type(sample_df_pandas, tmpdir):
    filepath = os.path.join(tmpdir, 'sample')
    sample_df_pandas.to_csv(filepath, index=False)

    no_type_error = "Content type could not be inferred. Please specify content_type and try again."
    with pytest.raises(RuntimeError, match=no_type_error):
        ww.read_file(filepath=filepath)


def test_read_file_errors_unsupported(sample_df_pandas, tmpdir):
    filepath = os.path.join(tmpdir, 'sample')
    sample_df_pandas.to_feather(filepath)

    content_type = "notacontenttype"
    not_supported_error = "Reading from content type {} is not currently supported".format(content_type)
    with pytest.raises(RuntimeError, match=not_supported_error):
        ww.read_file(filepath=filepath, content_type=content_type)


@patch("woodwork.table_accessor._validate_accessor_params")
def test_read_file_validation_control(mock_validate_accessor_params, sample_df_pandas, tmpdir):
    filepath = os.path.join(tmpdir, 'sample.csv')
    sample_df_pandas.to_csv(filepath, index=False)

    assert not mock_validate_accessor_params.called
    ww.read_file(filepath=filepath, validate=False)
    assert not mock_validate_accessor_params.called
    ww.read_file(filepath=filepath)
    assert mock_validate_accessor_params.called


@pytest.mark.parametrize(
    "filepath, exportfn, kwargs, pandas_nullable_fix",
    [
        ("sample.csv", ("to_csv", {"index": False}), {}, True),
        ("sample.csv", ("to_csv", {"index": False}), {"content_type": "csv"}, True),
        ("sample.csv", ("to_csv", {"index": False}), {
            "index": 'id',
            "time_index": 'signup_date',
            "logical_types": {
                'full_name': 'NaturalLanguage',
                'phone_number': 'PhoneNumber',
                'is_registered': 'BooleanNullable',
                'age': 'IntegerNullable',
                'nullable_integer': 'IntegerNullable'},
            "semantic_tags": {
                'age': ['tag1', 'tag2'],
                'is_registered': ['tag3', 'tag4']
            },
            'column_origins': {
                'full_name': 'csv',
                'phone_number': 'base',
                'is_registered': 'engineered'
            }
        }, False),
        ("sample.csv", ("to_csv", {"index": False}), {"nrows": 2, "dtype": {'age': 'Int64', 'is_registered': 'boolean', 'nullable_integer': 'Int64'}}, False),
        ("sample.feather", ("to_feather", {}), {}, False),
        ("sample.feather", ("to_feather", {}), {"content_type": 'feather', "index": "id"}, False),
        ("sample.feather", ("to_feather", {}), {"content_type": 'application/feather', "index": "id"}, False),
        ("sample.arrow", ("to_feather", {}), {}, False),
        ("sample.arrow", ("to_feather", {}), {"content_type": 'arrow', "index": "id"}, False),
        ("sample.arrow", ("to_feather", {}), {"content_type": 'application/arrow', "index": "id"}, False),
        ("sample.parquet", ("to_parquet", {}), {}, False),
        ("sample.parquet", ("to_parquet", {}), {"content_type": 'parquet', "index": "id", "use_nullable_dtypes": True}, False),
        ("sample.parquet", ("to_parquet", {}), {"content_type": 'application/parquet', "index": "id", "use_nullable_dtypes": True}, False),
        ("sample.orc", (save_orc_file, {}), {}, True),
        ("sample.orc", (save_orc_file, {}), {"content_type": 'orc', "index": "id"}, True),
        ("sample.orc", (save_orc_file, {}), {"content_type": 'application/orc', "index": "id"}, True),
    ]
)
def test_read_file(sample_df_pandas, tmpdir, filepath, exportfn, kwargs, pandas_nullable_fix):
    filepath = os.path.join(tmpdir, filepath)
    func, func_kwargs = exportfn
    if isinstance(func, str):
        getattr(sample_df_pandas, func)(filepath, **func_kwargs)
    else:
        # Call save_orc_file to save orc data since pandas does not have a to_orc method
        func(sample_df_pandas, filepath, **func_kwargs)
    df = ww.read_file(filepath=filepath, **kwargs)
    assert isinstance(df.ww.schema, ww.table_schema.TableSchema)

    schema_df = sample_df_pandas.copy()
    if pandas_nullable_fix:
        # pandas does not read data into nullable types currently from csv or orc,
        # so the types in df will be different than the types inferred from sample_df_pandas
        # which uses the nullable types
        schema_df = schema_df.astype({'age': 'float64', 'nullable_integer': 'float64', 'is_registered': 'object'})

    if func in ["to_csv", save_orc_file]:  # categorical column not inferred as categorical
        schema_df['categorical'] = schema_df['categorical'].astype('string')

    schema_df.ww.init(index=kwargs.get('index'),
                      time_index=kwargs.get('time_index'),
                      logical_types=kwargs.get('logical_types'),
                      semantic_tags=kwargs.get('semantic_tags'),
                      column_origins=kwargs.get('column_origins'))

    if func == "to_csv":
        for c in ['signup_date', 'datetime_with_NaT']:
            df.ww.logical_types[c].datetime_format = None  # read_csv reads datetimes as strings and infers datetime during transform

    assert df.ww.schema == schema_df.ww.schema

    if "nrows" in kwargs:
        assert len(df) == kwargs["nrows"]
        schema_df = schema_df.head(kwargs["nrows"])

    pd.testing.assert_frame_equal(df, schema_df)



import json
import os

import boto3
import pandas as pd
import pytest
from mock import patch

import woodwork.deserialize as deserialize
import woodwork.serialize as serialize
from woodwork.accessor_utils import _is_dask_dataframe, _is_koalas_dataframe
from woodwork.exceptions import (
    OutdatedSchemaWarning,
    UpgradeSchemaWarning,
    WoodworkNotInitError
)
from woodwork.logical_types import Categorical, Ordinal
from woodwork.tests.testing_utils import to_pandas

BUCKET_NAME = "test-bucket"
WRITE_KEY_NAME = "test-key"
TEST_S3_URL = "s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)
TEST_FILE = "test_serialization_woodwork_table_schema_{}.tar".format(serialize.SCHEMA_VERSION)
S3_URL = "s3://woodwork-static/" + TEST_FILE
URL = "https://woodwork-static.s3.amazonaws.com/" + TEST_FILE
TEST_KEY = "test_access_key_es"


def xfail_tmp_disappears(dataframe):
    # TODO: tmp file disappears after deserialize step, cannot check equality with Dask
    if not isinstance(dataframe, pd.DataFrame):
        pytest.xfail('tmp file disappears after deserialize step, cannot check equality with Dask')


def test_error_before_table_init(sample_df, tmpdir):
    error_message = "Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init"

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.to_dictionary()

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.to_disk(str(tmpdir), format='csv')


def test_to_dictionary(sample_df):
    if _is_dask_dataframe(sample_df):
        table_type = 'dask'
        age_cat_type_dict = {
            'type': 'category',
            'cat_values': [33, 57],
            'cat_dtype': 'int64'
        }
        cat_type_dict = {
            'type': 'category',
            'cat_values': ['a', 'b', 'c'],
            'cat_dtype': 'object'
        }
    elif _is_koalas_dataframe(sample_df):
        table_type = 'koalas'
        age_cat_type_dict = {
            'type': 'string'
        }
        cat_type_dict = {
            'type': 'string'
        }
    else:
        table_type = 'pandas'
        age_cat_type_dict = {
            'type': 'category',
            'cat_values': [33, 57],
            'cat_dtype': 'object'
        }
        cat_type_dict = {
            'type': 'category',
            'cat_values': ['a', 'b', 'c'],
            'cat_dtype': 'object'
        }

    int_val = 'int64'
    nullable_int_val = 'Int64'
    string_val = 'string'
    bool_val = 'boolean'
    double_val = 'float64'

    expected = {'schema_version': serialize.SCHEMA_VERSION,
                'name': 'test_data',
                'index': 'id',
                'time_index': None,
                'column_typing_info': [{'name': 'id',
                                        'ordinal': 0,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Integer'},
                                        'physical_type': {'type': int_val},
                                        'semantic_tags': ['index', 'tag1'],
                                        'description': None,
                                        'origin': None,
                                        'metadata':{'is_sorted': True}},
                                       {'name': 'full_name',
                                        'ordinal': 1,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Unknown'},
                                        'physical_type': {'type': string_val},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': None,
                                        'metadata':{}},
                                       {'name': 'email',
                                        'ordinal': 2,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'EmailAddress'},
                                        'physical_type': {'type': string_val},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': None,
                                        'metadata':{}},
                                       {'name': 'phone_number',
                                        'ordinal': 3,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Unknown'},
                                        'physical_type': {'type': string_val},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': 'base',
                                        'metadata': {}},
                                       {'name': 'age',
                                        'ordinal': 4,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {'order': [25, 33, 57]}, 'type': 'Ordinal'},
                                        'physical_type': age_cat_type_dict,
                                        'semantic_tags': ['category'],
                                        'description': 'age of the user',
                                        'origin': 'base',
                                        'metadata':{'interesting_values': [33, 57]}},
                                       {'name': 'signup_date',
                                        'ordinal': 5,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {'datetime_format': None},
                                                         'type': 'Datetime'},
                                        'physical_type': {'type': 'datetime64[ns]'},
                                        'semantic_tags': [],
                                        'description': 'original signup date',
                                        'origin': 'engineered',
                                        'metadata':{}},
                                       {'name': 'is_registered',
                                        'ordinal': 6,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'BooleanNullable'},
                                        'physical_type': {'type': bool_val},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'double',
                                        'ordinal': 7,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Double'},
                                        'physical_type': {'type': double_val},
                                        'semantic_tags': ['numeric'],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'double_with_nan',
                                        'ordinal': 8,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Double'},
                                        'physical_type': {'type': double_val},
                                        'semantic_tags': ['numeric'],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'integer',
                                        'ordinal': 9,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Integer'},
                                        'physical_type': {'type': int_val},
                                        'semantic_tags': ['numeric'],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'nullable_integer',
                                        'ordinal': 10,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'IntegerNullable'},
                                        'physical_type': {'type': nullable_int_val},
                                        'semantic_tags': ['numeric'],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'boolean',
                                        'ordinal': 11,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Boolean'},
                                        'physical_type': {'type': 'bool'},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'categorical',
                                        'ordinal': 12,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {}, 'type': 'Categorical'},
                                        'physical_type': cat_type_dict,
                                        'semantic_tags': ['category'],
                                        'description': None,
                                        'origin': None,
                                        'metadata': {}},
                                       {'name': 'datetime_with_NaT',
                                        'ordinal': 13,
                                        'use_standard_tags': True,
                                        'logical_type': {'parameters': {'datetime_format': None}, 'type': 'Datetime'},
                                        'physical_type': {'type': 'datetime64[ns]'},
                                        'semantic_tags': [],
                                        'description': None,
                                        'origin': None,
                                        'metadata':{}}],
                'loading_info': {'table_type': table_type},
                'table_metadata': {'date_created': '11/16/20'}
                }
    sample_df.ww.init(
        name='test_data',
        index='id',
        semantic_tags={'id': 'tag1'},
        logical_types={'age': Ordinal(order=[25, 33, 57])},
        table_metadata={'date_created': '11/16/20'},
        column_descriptions={'signup_date': 'original signup date',
                             'age': 'age of the user'},
        column_origins={'phone_number': 'base',
                        'age': 'base',
                        'signup_date': 'engineered'},
        column_metadata={'id': {'is_sorted': True},
                         'age': {'interesting_values': [33, 57]}}
    )

    description = sample_df.ww.to_dictionary()
    assert description == expected


def test_unserializable_table(sample_df, tmpdir):
    sample_df.ww.init(table_metadata={'not_serializable': sample_df['is_registered'].dtype})

    error = "Woodwork table is not json serializable. Check table and column metadata for values that may not be serializable."
    with pytest.raises(TypeError, match=error):
        sample_df.ww.to_disk(str(tmpdir), format='csv', encoding='utf-8', engine='python')


def test_serialize_wrong_format(sample_df, tmpdir):
    sample_df.ww.init()

    error = 'must be one of the following formats: csv, pickle, parquet'
    with pytest.raises(ValueError, match=error):
        serialize.write_woodwork_table(sample_df, str(tmpdir), format='test')


def test_to_csv(sample_df, tmpdir):
    if _is_dask_dataframe(sample_df):
        # Dask errors with pd.NA in some partitions, but not others
        sample_df['age'] = sample_df['age'].fillna(25)
    sample_df.ww.init(
        name='test_data',
        index='id',
        semantic_tags={'id': 'tag1'},
        logical_types={'age': Ordinal(order=[25, 33, 57])},
        column_descriptions={'signup_date': 'original signup date',
                             'age': 'age of the user'},
        column_origins={'phone_number': 'base',
                        'age': 'base',
                        'signup_date': 'engineered'},
        column_metadata={'id': {'is_sorted': True},
                         'age': {'interesting_values': [33, 57]}})
    sample_df.ww.to_disk(str(tmpdir), format='csv', encoding='utf-8', engine='python')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
                                  to_pandas(sample_df, index=sample_df.ww.index, sort_index=True))
    assert deserialized_df.ww.schema == sample_df.ww.schema


def test_to_csv_with_latlong(latlong_df, tmpdir):
    latlong_df.ww.init(index='tuple_ints', logical_types={col: 'LatLong' for col in latlong_df.columns})
    latlong_df.ww.to_disk(str(tmpdir))
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
                                  to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True))
    assert deserialized_df.ww.schema == latlong_df.ww.schema


def test_to_disk_with_whitespace(whitespace_df, tmpdir):
    formats = ['csv', 'parquet', 'pickle']
    for format in formats:
        df = whitespace_df.copy()
        df.ww.init(index='id', logical_types={'comments': 'NaturalLanguage'})
        if format == 'pickle' and not isinstance(df, pd.DataFrame):
            msg = 'DataFrame type not compatible with pickle serialization. Please serialize to another format.'
            with pytest.raises(ValueError, match=msg):
                df.ww.to_disk(str(tmpdir), format='pickle')
        else:
            df.ww.to_disk(str(tmpdir), format=format)
            deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
            assert deserialized_df.ww.schema == df.ww.schema
            pd.testing.assert_frame_equal(to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
                                          to_pandas(df, index=df.ww.index, sort_index=True))


def test_to_csv_use_standard_tags(sample_df, tmpdir):
    no_standard_tags_df = sample_df.copy()
    no_standard_tags_df.ww.init(use_standard_tags=False)

    no_standard_tags_df.ww.to_disk(str(tmpdir), format='csv', encoding='utf-8', engine='python')
    deserialized_no_tags_df = deserialize.read_woodwork_table(str(tmpdir))

    standard_tags_df = sample_df.copy()
    standard_tags_df.ww.init(use_standard_tags=True)

    standard_tags_df.ww.to_disk(str(tmpdir), format='csv', encoding='utf-8', engine='python')
    deserialized_tags_df = deserialize.read_woodwork_table(str(tmpdir))

    assert no_standard_tags_df.ww.schema != standard_tags_df.ww.schema

    assert deserialized_no_tags_df.ww.schema == no_standard_tags_df.ww.schema
    assert deserialized_tags_df.ww.schema == standard_tags_df.ww.schema


def test_to_pickle(sample_df, tmpdir):
    sample_df.ww.init()
    if not isinstance(sample_df, pd.DataFrame):
        msg = 'DataFrame type not compatible with pickle serialization. Please serialize to another format.'
        with pytest.raises(ValueError, match=msg):
            sample_df.ww.to_disk(str(tmpdir), format='pickle')
    else:
        sample_df.ww.to_disk(str(tmpdir), format='pickle')
        deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

        pd.testing.assert_frame_equal(to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
                                      to_pandas(sample_df, index=sample_df.ww.index, sort_index=True))
        assert deserialized_df.ww.schema == sample_df.ww.schema


def test_to_pickle_with_latlong(latlong_df, tmpdir):
    latlong_df.ww.init(logical_types={col: 'LatLong' for col in latlong_df.columns})
    if not isinstance(latlong_df, pd.DataFrame):
        msg = 'DataFrame type not compatible with pickle serialization. Please serialize to another format.'
        with pytest.raises(ValueError, match=msg):
            latlong_df.ww.to_disk(str(tmpdir), format='pickle')
    else:
        latlong_df.ww.to_disk(str(tmpdir), format='pickle')
        deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

        pd.testing.assert_frame_equal(to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True),
                                      to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
        assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_to_parquet(sample_df, tmpdir):
    sample_df.ww.init(index='id')
    sample_df.ww.to_disk(str(tmpdir), format='parquet')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_to_parquet_with_latlong(latlong_df, tmpdir):
    latlong_df.ww.init(logical_types={col: 'LatLong' for col in latlong_df.columns})
    latlong_df.ww.to_disk(str(tmpdir), format='parquet')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_to_arrow(sample_df, tmpdir):
    if not isinstance(sample_df, pd.DataFrame):
        pytest.xfail('Arrow IPC format (Feather) not supported on Dask or Koalas')
    sample_df.ww.init(index='id')
    sample_df.ww.to_disk(str(tmpdir), format='arrow')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_to_arrow_with_latlong(latlong_df, tmpdir):
    if not isinstance(latlong_df, pd.DataFrame):
        pytest.xfail('Arrow IPC format (Feather) not supported on Dask or Koalas')
    latlong_df.ww.init(logical_types={col: 'LatLong' for col in latlong_df.columns})
    latlong_df.ww.to_disk(str(tmpdir), format='arrow')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_to_feather(sample_df, tmpdir):
    if not isinstance(sample_df, pd.DataFrame):
        pytest.xfail('Arrow IPC format (Feather) not supported on Dask or Koalas')
    sample_df.ww.init(index='id')
    sample_df.ww.to_disk(str(tmpdir), format='feather')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_to_feather_with_latlong(latlong_df, tmpdir):
    if not isinstance(latlong_df, pd.DataFrame):
        pytest.xfail('Arrow IPC format (Feather) not supported on Dask or Koalas')
    latlong_df.ww.init(logical_types={col: 'LatLong' for col in latlong_df.columns})
    latlong_df.ww.to_disk(str(tmpdir), format='feather')
    deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_to_orc(sample_df, tmpdir):
    sample_df.ww.init(index='id')
    if _is_dask_dataframe(sample_df):
        msg = 'DataFrame type not compatible with orc serialization. Please serialize to another format.'
        with pytest.raises(ValueError, match=msg):
            sample_df.ww.to_disk(str(tmpdir), format='orc')
    else:
        sample_df.ww.to_disk(str(tmpdir), format='orc')
        deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
        pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                      to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
        assert sample_df.ww.schema == deserialized_df.ww.schema


def test_to_orc_with_latlong(latlong_df, tmpdir):
    latlong_df.ww.init(logical_types={col: 'LatLong' for col in latlong_df.columns})
    if _is_dask_dataframe(latlong_df):
        msg = 'DataFrame type not compatible with orc serialization. Please serialize to another format.'
        with pytest.raises(ValueError, match=msg):
            latlong_df.ww.to_disk(str(tmpdir), format='orc')
    else:
        latlong_df.ww.to_disk(str(tmpdir), format='orc')
        deserialized_df = deserialize.read_woodwork_table(str(tmpdir))

        pd.testing.assert_frame_equal(to_pandas(latlong_df, index=latlong_df.ww.index, sort_index=True),
                                      to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
        assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_categorical_dtype_serialization(serialize_df, tmpdir):
    ltypes = {
        'cat_int': Categorical,
        'ord_int': Ordinal(order=[1, 2]),
        'cat_float': Categorical,
        'ord_float': Ordinal(order=[1.0, 2.0]),
        'cat_bool': Categorical,
        'ord_bool': Ordinal(order=[True, False]),
    }
    if isinstance(serialize_df, pd.DataFrame):
        formats = ['csv', 'pickle', 'parquet']
    else:
        formats = ['csv']

    for format in formats:
        df = serialize_df.copy()
        df.ww.init(index='id', logical_types=ltypes)
        df.ww.to_disk(str(tmpdir), format=format)
        deserialized_df = deserialize.read_woodwork_table(str(tmpdir))
        pd.testing.assert_frame_equal(to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
                                      to_pandas(df, index=df.ww.index, sort_index=True))
        assert deserialized_df.ww.schema == df.ww.schema


@pytest.fixture
def s3_client():
    # TODO: Fix Moto tests needing to explicitly set permissions for objects
    _environ = os.environ.copy()
    from moto import mock_s3
    with mock_s3():
        s3 = boto3.resource('s3')
        yield s3
    os.environ.clear()
    os.environ.update(_environ)


@pytest.fixture
def s3_bucket(s3_client):
    s3_client.create_bucket(Bucket=BUCKET_NAME, ACL='public-read-write')
    s3_bucket = s3_client.Bucket(BUCKET_NAME)
    yield s3_bucket


def make_public(s3_client, s3_bucket):
    obj = list(s3_bucket.objects.all())[0].key
    s3_client.ObjectAcl(BUCKET_NAME, obj).put(ACL='public-read-write')


def test_to_csv_S3(sample_df, s3_client, s3_bucket):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init(
        name='test_data',
        index='id',
        semantic_tags={'id': 'tag1'},
        logical_types={'age': Ordinal(order=[25, 33, 57])})
    sample_df.ww.to_disk(TEST_S3_URL, format='csv', encoding='utf-8', engine='python')
    make_public(s3_client, s3_bucket)

    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL)

    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_serialize_s3_pickle(sample_df_pandas, s3_client, s3_bucket):
    sample_df_pandas.ww.init()
    sample_df_pandas.ww.to_disk(TEST_S3_URL, format='pickle')
    make_public(s3_client, s3_bucket)
    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL)

    pd.testing.assert_frame_equal(to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


def test_serialize_s3_parquet(sample_df, s3_client, s3_bucket):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init()
    sample_df.ww.to_disk(TEST_S3_URL, format='parquet')
    make_public(s3_client, s3_bucket)
    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL)

    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_to_csv_S3_anon(sample_df, s3_client, s3_bucket):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init(
        name='test_data',
        index='id',
        time_index='signup_date',
        semantic_tags={'id': 'tag1'},
        logical_types={'age': Ordinal(order=[25, 33, 57])})
    sample_df.ww.to_disk(TEST_S3_URL, format='csv', encoding='utf-8', engine='python', profile_name=False)
    make_public(s3_client, s3_bucket)

    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL, profile_name=False)

    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
                                  to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_serialize_s3_pickle_anon(sample_df_pandas, s3_client, s3_bucket):
    sample_df_pandas.ww.init()
    sample_df_pandas.ww.to_disk(TEST_S3_URL, format='pickle', profile_name=False)
    make_public(s3_client, s3_bucket)
    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL, profile_name=False)

    pd.testing.assert_frame_equal(to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


def test_serialize_s3_parquet_anon(sample_df, s3_client, s3_bucket):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init()
    sample_df.ww.to_disk(TEST_S3_URL, format='parquet', profile_name=False)
    make_public(s3_client, s3_bucket)
    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL, profile_name=False)

    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def create_test_credentials(test_path):
    with open(test_path, "w+") as f:
        f.write("[test]\n")
        f.write("aws_access_key_id=AKIAIOSFODNN7EXAMPLE\n")
        f.write("aws_secret_access_key=wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY\n")


def create_test_config(test_path_config):
    with open(test_path_config, "w+") as f:
        f.write("[profile test]\n")
        f.write("region=us-east-2\n")
        f.write("output=text\n")


@pytest.fixture
def setup_test_profile(monkeypatch, tmpdir):
    cache = str(tmpdir.join('.cache').mkdir())
    test_path = os.path.join(cache, 'test_credentials')
    test_path_config = os.path.join(cache, 'test_config')
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", test_path)
    monkeypatch.setenv("AWS_CONFIG_FILE", test_path_config)
    monkeypatch.delenv("AWS_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("AWS_SECRET_ACCESS_KEY", raising=False)
    monkeypatch.setenv("AWS_PROFILE", "test")

    try:
        os.remove(test_path)
    except OSError:
        pass
    try:
        os.remove(test_path_config)
    except OSError:
        pass

    create_test_credentials(test_path)
    create_test_config(test_path_config)
    yield
    os.remove(test_path)
    os.remove(test_path_config)


def test_s3_test_profile(sample_df, s3_client, s3_bucket, setup_test_profile):
    xfail_tmp_disappears(sample_df)
    sample_df.ww.init()
    sample_df.ww.to_disk(TEST_S3_URL, format='csv', encoding='utf-8', engine='python', profile_name='test')
    make_public(s3_client, s3_bucket)
    deserialized_df = deserialize.read_woodwork_table(TEST_S3_URL, profile_name='test')

    pd.testing.assert_frame_equal(to_pandas(sample_df, index=sample_df.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_serialize_url_csv(sample_df):
    sample_df.ww.init()
    error_text = "Writing to URLs is not supported"
    with pytest.raises(ValueError, match=error_text):
        sample_df.ww.to_disk(URL, format='csv', encoding='utf-8', engine='python')


def test_serialize_subdirs_not_removed(sample_df, tmpdir):
    sample_df.ww.init()
    write_path = tmpdir.mkdir("test")
    test_dir = write_path.mkdir("test_dir")
    with open(str(write_path.join('woodwork_typing_info.json')), 'w') as f:
        json.dump('__SAMPLE_TEXT__', f)
    compression = None
    serialize.write_woodwork_table(sample_df, path=str(write_path), index='1', sep='\t', encoding='utf-8', compression=compression)
    assert os.path.exists(str(test_dir))
    with open(str(write_path.join('woodwork_typing_info.json')), 'r') as f:
        assert '__SAMPLE_TEXT__' not in json.load(f)


def test_deserialize_url_csv(sample_df_pandas):
    sample_df_pandas.ww.init(index='id')
    deserialized_df = deserialize.read_woodwork_table(URL)
    pd.testing.assert_frame_equal(to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


def test_deserialize_url_csv_anon(sample_df_pandas):
    sample_df_pandas.ww.init(index='id')
    deserialized_df = deserialize.read_woodwork_table(URL, profile_name=False)

    pd.testing.assert_frame_equal(to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


def test_deserialize_s3_csv(sample_df_pandas):
    sample_df_pandas.ww.init(index='id')
    deserialized_df = deserialize.read_woodwork_table(S3_URL)

    pd.testing.assert_frame_equal(to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index), to_pandas(deserialized_df, index=deserialized_df.ww.index))
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


@patch("woodwork.table_accessor._validate_accessor_params")
def test_deserialize_validation_control(mock_validate_accessor_params):
    assert not mock_validate_accessor_params.called
    deserialize.read_woodwork_table(URL)
    assert not mock_validate_accessor_params.called
    deserialize.read_woodwork_table(URL, validate=True)
    assert mock_validate_accessor_params.called


def test_check_later_schema_version():
    def test_version(major, minor, patch, raises=True):
        version_to_check = '.'.join([str(v) for v in [major, minor, patch]])
        if raises:
            warning_text = ('The schema version of the saved Woodwork table '
                            '%s is greater than the latest supported %s. '
                            'You may need to upgrade woodwork. Attempting to load Woodwork table ...'
                            % (version_to_check, serialize.SCHEMA_VERSION))
            with pytest.warns(UpgradeSchemaWarning, match=warning_text):
                deserialize._check_schema_version(version_to_check)
        else:
            with pytest.warns(None) as record:
                deserialize._check_schema_version(version_to_check)
            assert len(record) == 0

    major, minor, patch = [int(s) for s in serialize.SCHEMA_VERSION.split('.')]

    test_version(major + 1, minor, patch)
    test_version(major, minor + 1, patch)
    test_version(major, minor, patch + 1)
    test_version(major, minor - 1, patch + 1, raises=False)


def test_earlier_schema_version():
    def test_version(major, minor, patch, raises=True):
        version_to_check = '.'.join([str(v) for v in [major, minor, patch]])
        if raises:
            warning_text = ('The schema version of the saved Woodwork table '
                            '%s is no longer supported by this version '
                            'of woodwork. Attempting to load Woodwork table ...'
                            % (version_to_check))
            with pytest.warns(OutdatedSchemaWarning, match=warning_text):
                deserialize._check_schema_version(version_to_check)
        else:
            with pytest.warns(None) as record:
                deserialize._check_schema_version(version_to_check)
            assert len(record) == 0

    major, minor, patch = [int(s) for s in serialize.SCHEMA_VERSION.split('.')]

    test_version(major - 1, minor, patch)
    test_version(major, minor - 1, patch, raises=False)
    test_version(major, minor, patch - 1, raises=False)



import pandas as pd
import pytest
from mock import patch

import woodwork as ww
from woodwork.accessor_utils import _is_koalas_dataframe
from woodwork.logical_types import Categorical, Double, Integer
from woodwork.tests.testing_utils import to_pandas
from woodwork.utils import concat_columns


def test_concat_cols_ww_dfs(sample_df):
    sample_df.ww.init(logical_types={'full_name': 'Categorical'}, semantic_tags={'age': 'test_tag'},
                      table_metadata={'created_by': 'user0'})
    df1 = sample_df.ww[['id', 'full_name', 'email', 'phone_number', 'age', 'signup_date', 'is_registered']]
    df2 = sample_df.ww[['double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    df2.ww.metadata = None

    combined_df = concat_columns([df1, df2])
    assert combined_df.ww == sample_df.ww
    assert isinstance(combined_df.ww.logical_types['full_name'], Categorical)
    assert 'test_tag' in combined_df.ww.semantic_tags['age']
    assert combined_df.ww.metadata == {'created_by': 'user0'}

    pandas_combined_df = pd.concat([to_pandas(df1), to_pandas(df2)], axis=1, join='outer', ignore_index=False)
    assert to_pandas(combined_df).equals(pandas_combined_df)


def test_concat_cols_uninit_dfs(sample_df):
    df1 = sample_df[['id', 'full_name', 'email', 'phone_number', 'age', 'signup_date', 'is_registered']]
    df2 = sample_df[['double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    sample_df.ww.init()

    combined_df = concat_columns([df1, df2])
    assert df1.ww.schema is None
    assert df2.ww.schema is None
    assert combined_df.ww.schema is not None

    # Though the input dataframes don't have Woodwork initalized,
    # the concatenated DataFrame has Woodwork initialized for all the columns
    assert isinstance(combined_df.ww.logical_types['id'], Integer)
    assert combined_df.ww == sample_df.ww

    df1.ww.init()
    df2.ww.init()
    pandas_combined_df = pd.concat([to_pandas(df1), to_pandas(df2)], axis=1, join='outer', ignore_index=False)
    assert to_pandas(combined_df).equals(pandas_combined_df)


def test_concat_cols_combo_dfs(sample_df):
    df1 = sample_df[['id', 'full_name', 'email', 'phone_number', 'age', 'signup_date', 'is_registered']]

    sample_df.ww.init()
    df2 = sample_df.ww[['double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]

    combined_df = concat_columns([df1, df2])
    assert df1.ww.schema is None
    assert combined_df.ww == sample_df.ww

    df1.ww.init()
    df2.ww.init()
    pandas_combined_df = pd.concat([to_pandas(df1), to_pandas(df2)], axis=1, join='outer', ignore_index=False)
    assert to_pandas(combined_df).equals(pandas_combined_df)


def test_concat_cols_with_series(sample_df):
    expected_df = sample_df[['id', 'full_name', 'signup_date', 'is_registered']]
    df = expected_df[['id', 'full_name']]
    s1 = expected_df['signup_date']
    expected_df.ww.init()
    s2 = expected_df['is_registered']

    combined_df = concat_columns([df, s1, s2])
    assert combined_df.ww == expected_df.ww

    df.ww.init()
    s1.ww.init()
    s2.ww.init()
    pandas_combined_df = pd.concat([to_pandas(df), to_pandas(s1), to_pandas(s2)], axis=1, join='outer', ignore_index=False)
    assert to_pandas(combined_df).equals(pandas_combined_df)


def test_concat_cols_with_conflicting_ww_indexes(sample_df):
    df1 = sample_df[['id', 'full_name', 'email', 'phone_number', 'age', 'signup_date', 'is_registered']]
    df1.ww.init(index='id')
    df2 = sample_df[['double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    df2.ww.init(index='double')

    error = ('Cannot set the Woodwork index of multiple input objects. '
             'Please remove the index columns from all but one table.')

    with pytest.raises(IndexError, match=error):
        concat_columns([df1, df2])

    df1 = sample_df[['id', 'full_name', 'email', 'phone_number', 'age', 'is_registered']]
    df1.ww.init(time_index='id')
    df2 = sample_df[['signup_date', 'double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    df2.ww.init(time_index='signup_date')

    error = ('Cannot set the Woodwork time index of multiple input objects. '
             'Please remove the time index columns from all but one table.')
    with pytest.raises(IndexError, match=error):
        concat_columns([df1, df2])


def test_concat_cols_with_ww_indexes(sample_df):
    df1 = sample_df[['id', 'full_name', 'email', 'phone_number', 'age', 'is_registered']]
    df1.ww.init(index='id')
    df2 = sample_df[['signup_date', 'double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    df2.ww.init(time_index='signup_date')

    combined_df = concat_columns([df1, df2])
    assert combined_df.ww.index == 'id'
    assert combined_df.ww.time_index == 'signup_date'


def test_concat_cols_with_duplicate_ww_indexes(sample_df):
    df1 = sample_df[['id', 'signup_date', 'full_name', 'email', 'phone_number', 'age', 'is_registered']]
    df1.ww.init(index='id', time_index='signup_date')
    df2 = sample_df[['id', 'signup_date', 'double', 'double_with_nan', 'integer', 'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']]
    df2.ww.init(index='id', time_index='signup_date')

    error = ('Cannot set the Woodwork index of multiple input objects. '
             'Please remove the index columns from all but one table.')
    with pytest.raises(IndexError, match=error):
        concat_columns([df1, df2])

    df2.ww.pop('id')

    error = ('Cannot set the Woodwork time index of multiple input objects. '
             'Please remove the time index columns from all but one table.')
    with pytest.raises(IndexError, match=error):
        concat_columns([df1, df2])

    df2.ww.pop('signup_date')

    # Because underlying index is set, this won't change concat operation
    pd.testing.assert_index_equal(to_pandas(df1.index), to_pandas(df2.index))

    combined_df = concat_columns([df1, df2])
    assert combined_df.ww.index == 'id'
    assert combined_df.ww.time_index == 'signup_date'
    assert len(combined_df.columns) == len(sample_df.columns)


def test_concat_table_names(sample_df):
    df1 = sample_df[['id', 'signup_date']]
    df1.ww.init(name=0)
    df2 = sample_df[['full_name', 'age']]
    df2.ww.init(name='1')
    df3 = sample_df[['phone_number', 'is_registered']]
    df3.ww.init()

    combined_df = concat_columns([df1, df2, df3])
    assert combined_df.ww.name == '0_1'

    combined_df = concat_columns([df3, df2, df1])
    assert combined_df.ww.name == '1_0'


def test_concat_cols_different_use_standard_tags(sample_df):
    sample_df.ww.init(use_standard_tags={'age': False, 'full_name': False, 'id': True},
                      logical_types={'full_name': 'Categorical'})

    assert sample_df.ww.semantic_tags['id'] == {'numeric'}
    assert sample_df.ww.semantic_tags['full_name'] == set()
    assert sample_df.ww.semantic_tags['age'] == set()

    df1 = sample_df.ww[['id', 'full_name', 'email']]
    df2 = sample_df.ww[['phone_number', 'age', 'signup_date', 'is_registered']]

    combined_df = concat_columns([df1, df2])

    assert combined_df.ww.semantic_tags['id'] == {'numeric'}
    assert combined_df.ww.semantic_tags['full_name'] == set()
    assert combined_df.ww.semantic_tags['age'] == set()
    assert isinstance(combined_df.ww.logical_types['full_name'], Categorical)
    assert not combined_df.ww.use_standard_tags['age']
    assert not combined_df.ww.use_standard_tags['full_name']
    assert combined_df.ww.use_standard_tags['id']


def test_concat_combine_metadatas(sample_df):
    df1 = sample_df[['id', 'full_name', 'email']]
    df1.ww.init(table_metadata={'created_by': 'user0', 'test_key': 'test_val1'},
                column_metadata={'id': {'interesting_values': [1, 2]}})
    df2 = sample_df[['phone_number', 'age', 'signup_date', 'is_registered']]
    df2.ww.init(table_metadata={'table_type': 'single', 'test_key': 'test_val2'},
                column_metadata={'age': {'interesting_values': [33]}})

    error = "Cannot resolve overlapping keys in table metadata: {'test_key'}"
    with pytest.raises(ValueError, match=error):
        concat_columns([df1, df2])

    del df2.ww.metadata['test_key']

    combined_df = concat_columns([df1, df2])

    assert combined_df.ww.metadata == {'created_by': 'user0',
                                       'table_type': 'single',
                                       'test_key': 'test_val1'}
    assert combined_df.ww.columns['id'].metadata == {'interesting_values': [1, 2]}
    assert combined_df.ww.columns['age'].metadata == {'interesting_values': [33]}


@patch("woodwork.table_accessor._validate_accessor_params")
def test_concat_cols_validate_schema(mock_validate_accessor_params, sample_df):
    df1 = sample_df[['id', 'full_name', 'email']]
    df1.ww.init(validate=False)
    df2 = sample_df[['phone_number', 'age', 'signup_date', 'is_registered']]
    df2.ww.init(validate=False)

    assert not mock_validate_accessor_params.called

    concat_columns([df1, df2], validate_schema=False)

    assert not mock_validate_accessor_params.called

    concat_columns([df1, df2], validate_schema=True)

    assert mock_validate_accessor_params.called


def test_concat_cols_mismatched_index_adds_single_nan(sample_df):
    sample_df.ww.init(index='id')

    df1 = sample_df.ww.loc[[0, 1, 2], ['id', 'full_name']]
    df2 = sample_df.ww.loc[[1, 2, 3], ['signup_date', 'email']]

    if isinstance(sample_df, pd.DataFrame):
        # Only pandas shows the dtype change because Dask and Koalas won't show until compute
        error = "Error converting datatype for id from type float64 to type int64. Please confirm the underlying data is consistent with logical type Integer."
        with pytest.raises(ww.exceptions.TypeConversionError, match=error):
            concat_columns([df1, df2])

    # If the dtype can handle nans, it won't change
    sample_df.ww.init(index='id', logical_types={'id': 'IntegerNullable'})

    df1 = sample_df.ww.loc[[0, 1, 2], ['id', 'full_name']]
    df2 = sample_df.ww.loc[[1, 2, 3], ['signup_date', 'email']]

    combined_df = concat_columns([df1, df2])
    assert len(combined_df) == 4


def test_concat_cols_mismatched_index_adds_multiple_nans(sample_df_pandas):
    # Only pandas checks for index uniqueness
    sample_df_pandas.ww.init(index='id', logical_types={'id': 'IntegerNullable'})

    df1 = sample_df_pandas.ww.loc[[0, 1], ['id', 'full_name']]
    df2 = sample_df_pandas.ww.loc[[2, 3], ['signup_date', 'email']]

    error = "Index column must be unique"
    with pytest.raises(IndexError, match=error):
        concat_columns([df1, df2])


def test_concat_cols_duplicate_columns(sample_df):
    sample_df.ww.init()
    df1 = sample_df.ww[['id', 'full_name', 'age']]
    df2 = sample_df.ww[['full_name', 'email', 'signup_date']]

    error = ("Duplicate column 'full_name' has been found in more than one input object. "
             "Please remove duplicate columns from all but one table.")
    with pytest.raises(ValueError, match=error):
        concat_columns([df1, df2])


def test_concat_cols_duplicate_columns_one_index(sample_df):
    sample_df.ww.init()
    df1 = sample_df.ww[['id', 'full_name', 'age']]
    df2 = sample_df.ww[['id', 'email', 'signup_date']]
    df2.ww.set_index('id')
    assert df1.ww.index is None
    assert df2.ww.index == 'id'

    error = ("Duplicate column 'id' has been found in more than one input object. "
             "Please remove duplicate columns from all but one table.")
    with pytest.raises(ValueError, match=error):
        concat_columns([df1, df2])


def test_concat_table_order(sample_df):
    df1 = sample_df[['signup_date']]
    df1.ww.init()
    df2 = sample_df[['id', 'age']]
    df2.ww.init(index='id')

    combined_df1 = concat_columns([df1.ww.copy(), df2.ww.copy()])
    combined_df2 = concat_columns([df2.ww.copy(), df1.ww.copy()])

    assert set(combined_df1.columns) == set(combined_df2.columns)

    assert list(combined_df1.columns) == ['signup_date', 'id', 'age']
    assert list(combined_df2.columns) == ['id', 'age', 'signup_date']


def test_concat_cols_all_series(sample_df):
    age_series = ww.init_series(sample_df['age'], logical_type='Double')
    combined_df_1 = concat_columns([sample_df['id'], age_series, sample_df['is_registered']])
    combined_df_2 = concat_columns([age_series, sample_df['id'], sample_df['is_registered']])

    assert isinstance(combined_df_1.ww.schema, ww.table_schema.TableSchema)
    assert list(combined_df_1.columns) == ['id', 'age', 'is_registered']
    assert list(combined_df_2.columns) == ['age', 'id', 'is_registered']
    assert isinstance(combined_df_1.ww.logical_types['age'], Double)


def test_concat_cols_row_order(sample_df):
    sample_df.ww.init(index='id')
    pd.testing.assert_index_equal(to_pandas(sample_df.index), pd.Index([0, 1, 2, 3]))

    df1 = sample_df.ww.loc[:, ['id', 'full_name']]
    df2 = sample_df.ww.loc[[2, 3, 0, 1], ['email', 'phone_number']]
    df3 = sample_df.ww.loc[[3, 1, 0, 2], ['age',
                                          'signup_date',
                                          'is_registered',
                                          'double',
                                          'double_with_nan',
                                          'integer',
                                          'nullable_integer',
                                          'boolean',
                                          'categorical',
                                          'datetime_with_NaT']]

    combined_df = concat_columns([df1, df2, df3])

    assert sample_df.ww == combined_df.ww

    # koalas does not preserve index order in the same way
    if _is_koalas_dataframe(sample_df):
        pd.testing.assert_index_equal(to_pandas(combined_df.index), pd.Index([2, 0, 1, 3]))
    else:
        pd.testing.assert_frame_equal(to_pandas(sample_df), to_pandas(combined_df))


def test_concat_empty_list():
    error = 'No objects to concatenate'
    with pytest.raises(ValueError, match=error):
        concat_columns([])


def test_concat_with_none(sample_df):
    df1 = sample_df[['id', 'full_name', 'email']]
    df2 = sample_df[['phone_number', 'age', 'signup_date', 'is_registered']]

    error = "'NoneType' object has no attribute 'ww'"
    with pytest.raises(AttributeError, match=error):
        concat_columns([df1, df2, None])



import pandas as pd
import pytest

from woodwork.accessor_utils import _is_dask_dataframe, _is_dask_series
from woodwork.exceptions import WoodworkNotInitError
from woodwork.indexers import _iLocIndexer, _locIndexer
from woodwork.logical_types import (
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    PhoneNumber
)
from woodwork.tests.testing_utils import to_pandas
from woodwork.utils import import_or_none

dd = import_or_none('dask.dataframe')


def test_iLocIndexer_class_error(sample_df_dask, sample_series_dask):
    with pytest.raises(TypeError, match="iloc is not supported for Dask DataFrames"):
        _iLocIndexer(sample_df_dask)

    with pytest.raises(TypeError, match="iloc is not supported for Dask Series"):
        _iLocIndexer(sample_series_dask)


def test_iLocIndexer_class(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    sample_df.ww.init()
    ind = _iLocIndexer(sample_df)
    pd.testing.assert_frame_equal(to_pandas(ind.data), to_pandas(sample_df))
    pd.testing.assert_frame_equal(to_pandas(ind[1:2]), to_pandas(sample_df.iloc[1:2]))
    assert ind[0, 0] == 0


def test_locIndexer_class(sample_df):
    sample_df.ww.init()
    ind = _locIndexer(sample_df)
    pd.testing.assert_frame_equal(to_pandas(ind.data), to_pandas(sample_df))
    pd.testing.assert_frame_equal(to_pandas(ind[1:2]), to_pandas(sample_df.loc[1:2]))
    single_val = ind[0, 'id']
    if _is_dask_series(single_val):
        # Dask returns a series - convert to pandas to check the value
        single_val = single_val.compute()
        assert len(single_val) == 1
        single_val = single_val.loc[0]
    assert single_val == 0


def test_error_before_table_init(sample_df):
    error_message = "Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init"

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.iloc[:, :]

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.loc[:, :]


def test_error_before_column_init(sample_series):
    error_message = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_series.ww.iloc[:]

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_series.ww.loc[:]


def test_iloc_column(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail('iloc is not supported with Dask inputs')
    series = sample_series.copy()
    logical_type = Categorical
    semantic_tags = ['tag1', 'tag2']
    description = 'custom column description'
    origin = 'base'
    metadata = {'meta_key': 'custom metadata'}
    series.ww.init(logical_type=logical_type,
                   semantic_tags=semantic_tags,
                   description=description,
                   origin=origin,
                   metadata=metadata)

    sliced = series.ww.iloc[2:]
    assert sliced.name == "sample_series"
    assert isinstance(sliced.ww.logical_type, logical_type)
    assert sliced.ww.semantic_tags == {'category', 'tag1', 'tag2'}
    assert sliced.ww.description == description
    assert sliced.ww.origin == origin
    assert sliced.ww.metadata == metadata
    pd.testing.assert_series_equal(to_pandas(sliced), to_pandas(series.iloc[2:]))

    assert series.ww.iloc[0] == 'a'

    series = sample_series.copy()
    series.ww.init(use_standard_tags=False)
    sliced = series.ww.iloc[:]
    assert sliced.name
    assert isinstance(sliced.ww.logical_type, logical_type)
    assert sliced.ww.semantic_tags == set()


def test_iloc_column_does_not_propagate_changes_to_data(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail('iloc is not supported with Dask inputs')
    logical_type = Categorical
    semantic_tags = ['tag1', 'tag2']
    description = 'custom column description'
    origin = 'base'
    metadata = {'meta_key': 'custom metadata'}
    sample_series.ww.init(logical_type=logical_type,
                          semantic_tags=semantic_tags,
                          description=description,
                          origin=origin,
                          metadata=metadata,
                          use_standard_tags=False)

    sliced = sample_series.ww.iloc[2:]
    sample_series.ww.add_semantic_tags('new_tag')
    assert sliced.ww.semantic_tags == {'tag1', 'tag2'}
    assert sliced.ww.semantic_tags is not sample_series.ww.semantic_tags

    sample_series.ww.metadata['new_key'] = 'new_value'
    assert sliced.ww.metadata == {'meta_key': 'custom metadata'}
    assert sliced.ww.metadata is not sample_series.ww.metadata


def test_loc_column(sample_series):
    series = sample_series.copy()
    logical_type = Categorical
    semantic_tags = ['tag1', 'tag2']
    series.ww.init(logical_type=logical_type, semantic_tags=semantic_tags)

    sliced = series.ww.loc[2:]
    assert sliced.name == "sample_series"
    assert isinstance(sliced.ww.logical_type, logical_type)
    assert sliced.ww.semantic_tags == {'category', 'tag1', 'tag2'}
    pd.testing.assert_series_equal(to_pandas(sliced), to_pandas(series.loc[2:]))

    single_val = series.ww.loc[0]

    if _is_dask_series(series):
        # Dask returns a series - convert to pandas to check the value
        single_val = single_val.compute()
        assert len(single_val) == 1
        single_val = single_val.loc[0]
    assert single_val == 'a'

    series = sample_series.copy()
    series.ww.init(use_standard_tags=False)
    sliced = series.ww.loc[:]
    assert sliced.name
    assert isinstance(sliced.ww.logical_type, logical_type)
    assert sliced.ww.semantic_tags == set()


def test_iloc_indices_column(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    sample_df.ww.init(index='id', time_index='signup_date')
    sliced_index = sample_df.ww.iloc[:, 0]
    assert sliced_index.ww.semantic_tags == {'index'}

    sliced_time_index = sample_df.ww.iloc[:, 5]
    assert sliced_time_index.ww.semantic_tags == {'time_index'}


def test_loc_indices_column(sample_df):
    sample_df.ww.init(index='id', time_index='signup_date')
    sliced_index = sample_df.ww.loc[:, 'id']
    assert sliced_index.ww.semantic_tags == {'index'}

    sliced_time_index = sample_df.ww.loc[:, 'signup_date']
    assert sliced_time_index.ww.semantic_tags == {'time_index'}


def test_indexer_uses_standard_tags(sample_df):
    sample_df.ww.init(index='id', time_index='age', use_standard_tags={'id': False, 'age': True})
    sliced_index = sample_df.ww.loc[:, 'id']
    assert sliced_index.ww.semantic_tags == {'index'}

    sliced_time_index = sample_df.ww.loc[:, 'age']
    assert sliced_time_index.ww.semantic_tags == {'numeric', 'time_index'}


def test_iloc_with_properties(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'phone_number': ['tag3', 'tag2'],
        'signup_date': {'secondary_time_index'},
    }
    logical_types = {
        'full_name': Categorical,
        'email': EmailAddress,
        'phone_number': PhoneNumber,
        'age': Double,
    }
    df = sample_df.copy()
    df.ww.init(logical_types=logical_types, semantic_tags=semantic_tags)
    sliced = df.ww.iloc[1:3, 1:3]
    assert sliced.shape == (2, 2)
    assert sliced.ww.semantic_tags == {'full_name': {'category', 'tag1'}, 'email': {'tag2'}}
    assert isinstance(sliced.ww.logical_types['full_name'], Categorical)
    assert isinstance(sliced.ww.logical_types['email'], EmailAddress)
    assert sliced.ww.index is None

    df = sample_df.copy()
    df.ww.init(logical_types=logical_types, use_standard_tags=False)
    sliced = df.ww.iloc[:, [0, 5]]
    assert sliced.ww.semantic_tags == {'id': set(), 'signup_date': set()}
    assert isinstance(sliced.ww.logical_types['id'], Integer)
    assert isinstance(sliced.ww.logical_types['signup_date'], Datetime)
    assert sliced.ww.index is None


def test_iloc_dimensionality(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'phone_number': ['tag3', 'tag2'],
        'signup_date': {'secondary_time_index'},
    }
    logical_types = {
        'full_name': Categorical,
        'email': EmailAddress,
        'phone_number': PhoneNumber,
        'age': Double,
    }
    sample_df.ww.init(logical_types=logical_types, semantic_tags=semantic_tags)

    sliced_series_row = sample_df.ww.iloc[1]
    assert isinstance(sliced_series_row, pd.Series)
    assert set(sliced_series_row.index) == set(sample_df.columns)
    assert sliced_series_row.name == 1

    sliced_series_col = sample_df.ww.iloc[:, 1]
    assert isinstance(sliced_series_col.ww.logical_type, Categorical)
    assert sliced_series_col.ww.semantic_tags == {'tag1', 'category'}
    assert sliced_series_col.ww.name == 'full_name'


def test_iloc_indices(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    df_with_index = sample_df.copy()
    df_with_index.ww.init(index='id')
    assert df_with_index.ww.iloc[:, [0, 5]].ww.index == 'id'
    assert df_with_index.ww.iloc[:, [1, 2]].ww.index is None

    df_with_time_index = sample_df.copy()
    df_with_time_index.ww.init(time_index='signup_date')
    assert df_with_time_index.ww.iloc[:, [0, 5]].ww.time_index == 'signup_date'
    assert df_with_time_index.ww.iloc[:, [1, 2]].ww.index is None


def test_iloc_table_does_not_propagate_changes_to_data(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('iloc is not supported with Dask inputs')
    sample_df.ww.init()
    sliced = sample_df.ww.iloc[1:3, 1:3]

    sample_df.ww.add_semantic_tags({'full_name': 'new_tag'})
    assert sliced.ww.semantic_tags['full_name'] == set()
    assert sliced.ww.semantic_tags['full_name'] is not sample_df.ww.semantic_tags['full_name']

    sample_df.ww.metadata['new_key'] = 'new_value'
    assert sliced.ww.metadata == {}
    assert sliced.ww.metadata is not sample_df.ww.metadata

    sample_df.ww.columns['email'].metadata['new_key'] = 'new_value'
    assert sliced.ww.columns['email'].metadata == {}
    assert sliced.ww.columns['email'].metadata is not sample_df.ww.columns['email'].metadata



import urllib.request

import pytest

from woodwork.demo import load_retail
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage
)
from woodwork.table_schema import TableSchema


@pytest.fixture(autouse=True)
def set_testing_headers():
    opener = urllib.request.build_opener()
    opener.addheaders = [('Testing', 'True')]
    urllib.request.install_opener(opener)


def test_load_retail_diff():
    nrows = 10
    df = load_retail(nrows=nrows, init_woodwork=False)
    assert df.ww.schema is None
    assert df.shape[0] == nrows
    nrows_second = 11
    df = load_retail(nrows=nrows_second, init_woodwork=False)
    assert df.ww.schema is None
    assert df.shape[0] == nrows_second

    assert 'order_product_id' in df.columns
    assert df['order_product_id'].is_unique


def test_load_retail():
    df = load_retail(nrows=10, init_woodwork=True)
    assert isinstance(df.ww.schema, TableSchema)

    expected_logical_types = {
        'order_product_id': Categorical,
        'order_id': Categorical,
        'product_id': Categorical,
        'description': NaturalLanguage,
        'quantity': Integer,
        'order_date': Datetime,
        'unit_price': Double,
        'customer_name': Categorical,
        'country': Categorical,
        'total': Double,
        'cancelled': Boolean,
    }

    for col_name, column in df.ww.columns.items():
        assert isinstance(column.logical_type, expected_logical_types[col_name])

    assert df.ww.index == 'order_product_id'
    assert df.ww.time_index == 'order_date'
    assert df.ww.semantic_tags['order_product_id'] == {'index'}
    assert df.ww.semantic_tags['order_date'] == {'time_index'}



from datetime import datetime

import pandas as pd

import woodwork as ww
from woodwork.accessor_utils import _is_dask_series, _is_koalas_series
from woodwork.utils import import_or_none

ks = import_or_none('databricks.koalas')
dd = import_or_none('dask.dataframe')


def col_is_datetime(col, datetime_format=None):
    """Determine if a dataframe column contains datetime values or not. Returns True if column
    contains datetimes, False if not. Optionally specify the datetime format string for the column."""
    if _is_koalas_series(col):
        col = col.to_pandas()

    if (col.dtype.name.find('datetime') > -1 or
            (len(col) and isinstance(col.head(1), datetime))):
        return True

    # if it can be cast to numeric, it's not a datetime
    try:
        pd.to_numeric(col, errors='raise')
    except (ValueError, TypeError):
        # finally, try to cast to datetime
        if col.dtype.name.find('str') > -1 or col.dtype.name.find('object') > -1:
            try:
                pd.to_datetime(col, errors='raise', format=datetime_format, infer_datetime_format=True)
            except Exception:
                return False
            else:
                return True

    return False


def _is_numeric_series(series, logical_type):
    """Determines whether a series will be considered numeric
    for the purposes of determining if it can be a time_index."""
    if _is_koalas_series(series):
        series = series.to_pandas()
    if _is_dask_series(series):
        series = series.get_partition(0).compute()

    # If column can't be made to be numeric, don't bother checking Logical Type
    try:
        pd.to_numeric(series, errors='raise')
    except (ValueError, TypeError):
        return False

    if logical_type is not None:
        if isinstance(logical_type, str):
            logical_type = ww.type_system.str_to_logical_type(logical_type)

        # Allow numeric columns to be interpreted as Datetimes - doesn't allow strings even if they could be numeric
        if _get_ltype_class(logical_type) == ww.logical_types.Datetime and pd.api.types.is_numeric_dtype(series):
            return True
    else:
        logical_type = ww.type_system.infer_logical_type(series)

    return 'numeric' in logical_type.standard_tags


def list_logical_types():
    """Returns a dataframe describing all of the available Logical Types.

    Args:
        None

    Returns:
        pd.DataFrame: A dataframe containing details on each LogicalType, including
        the corresponding physical type and any standard semantic tags.
    """
    ltypes_df = pd.DataFrame(
        [{'name': ltype.__name__,
          'type_string': ltype.type_string,
          'description': ltype.__doc__,
          'physical_type': ltype.primary_dtype,
          'standard_tags': ltype.standard_tags,
          'is_default_type': ltype in ww.type_system._default_inference_functions,
          'is_registered': ltype in ww.type_system.registered_types,
          'parent_type': ww.type_system._get_parent(ltype)}
            for ltype in ww.logical_types.LogicalType.__subclasses__()]
    )
    return ltypes_df.sort_values('name').reset_index(drop=True)


def list_semantic_tags():
    """Returns a dataframe describing all of the common semantic tags.

    Args:
        None

    Returns:
        pd.DataFrame: A dataframe containing details on each Semantic Tag, including
        the corresponding logical type(s).
    """
    sem_tags = {}
    for ltype in ww.type_system.registered_types:
        for tag in ltype.standard_tags:
            if tag in sem_tags:
                sem_tags[tag].append(ltype)
            else:
                sem_tags[tag] = [ltype]
    tags_df = pd.DataFrame(
        [{'name': tag,
          'is_standard_tag': True,
          'valid_logical_types': sem_tags[tag]}
         for tag in sem_tags]
    )
    tags_df = tags_df.append(
        pd.DataFrame([['index', False, [ww.type_system.str_to_logical_type(tag) for tag in ['integer', 'double', 'categorical', 'datetime']]],
                      ['time_index', False, [ww.type_system.str_to_logical_type('datetime')]],
                      ['date_of_birth', False, [ww.type_system.str_to_logical_type('datetime')]]
                      ], columns=tags_df.columns), ignore_index=True)
    return tags_df


def _get_ltype_class(ltype):
    if ltype in ww.type_system.registered_types:
        return ltype
    return ltype.__class__


def _get_specified_ltype_params(ltype):
    """Gets a dictionary of a LogicalType's parameters.

    Note: If the logical type has not been instantiated, no parameters have
    been specified for the LogicalType, so no parameters will be returned
    even if that LogicalType can have parameters set.

    Args:
        ltype (LogicalType): An instantiated or uninstantiated LogicalType

    Returns:
        dict: The LogicalType's specified parameters.
    """
    if ltype in ww.type_system.registered_types:
        # Do not reveal parameters for an uninstantiated LogicalType
        return {}
    return ltype.__dict__


def _is_categorical_series(series: pd.Series, threshold: float) -> bool:
    """
    Return ``True`` if the given series is "likely" to be categorical.
    Otherwise, return ``False``.  We say that a series is "likely" to be
    categorical if the percentage of unique values relative to total non-NA
    values is below a certain threshold.  In other words, if all values in the
    series are accounted for by a sufficiently small collection of unique
    values, then the series is categorical.
    """
    try:
        nunique = series.nunique()
    except TypeError as e:
        # It doesn't seem like there's a more elegant way to do this.  Pandas
        # doesn't provide an API that would give you any indication ahead of
        # time if a series with object dtype has any unhashable elements.
        if "unhashable type" in e.args[0]:
            return False
        else:
            raise  # pragma: no cover
    if nunique == 0:
        return False

    pct_unique = nunique / series.count()
    return pct_unique <= threshold
